# pylint: disable=too-many-lines
import copy
import logging
from enum import Enum
from typing import Any, Dict, List, Optional

import dateutil
import more_itertools
import numpy
import pandas
from annofabapi.dataclass.statistics import ProjectAccountStatisticsHistory, WorktimeStatistics, WorktimeStatisticsItem
from annofabapi.models import (
    InputData,
    InputDataId,
    Inspection,
    InspectionStatus,
    Task,
    TaskHistory,
    TaskPhase,
    TaskStatus,
)
from annofabapi.utils import (
    get_number_of_rejections,
    get_task_history_index_skipped_acceptance,
    get_task_history_index_skipped_inspection,
)
from more_itertools import first_true

import annofabcli
from annofabcli.common.facade import AnnofabApiFacade
from annofabcli.common.utils import datetime_to_date, isoduration_to_hour
from annofabcli.statistics.database import AnnotationDict, Database

logger = logging.getLogger(__name__)


def _get_date_list(start_date: str, end_date: str) -> List[str]:
    return [e.strftime("%Y-%m-%d") for e in pandas.date_range(start=start_date, end=end_date)]


def _add_ratio_column_for_productivity_per_user(df: pandas.DataFrame, phase_list: List[str]):
    for phase in phase_list:
        # AnnoFab時間の比率
        df[("monitored_worktime_ratio", phase)] = (
            df[("monitored_worktime_hour", phase)] / df[("monitored_worktime_hour", "sum")]
        )
        # AnnoFab時間の比率から、Annowork時間を予測する
        df[("prediction_actual_worktime_hour", phase)] = (
            df[("actual_worktime_hour", "sum")] * df[("monitored_worktime_ratio", phase)]
        )

        # 生産性を算出
        df[("monitored_worktime/input_data_count", phase)] = (
            df[("monitored_worktime_hour", phase)] / df[("input_data_count", phase)]
        )
        df[("actual_worktime/input_data_count", phase)] = (
            df[("prediction_actual_worktime_hour", phase)] / df[("input_data_count", phase)]
        )

        df[("monitored_worktime/annotation_count", phase)] = (
            df[("monitored_worktime_hour", phase)] / df[("annotation_count", phase)]
        )
        df[("actual_worktime/annotation_count", phase)] = (
            df[("prediction_actual_worktime_hour", phase)] / df[("annotation_count", phase)]
        )

    phase = TaskPhase.ANNOTATION.value
    df[("pointed_out_inspection_comment_count/annotation_count", phase)] = (
        df[("pointed_out_inspection_comment_count", phase)] / df[("annotation_count", phase)]
    )
    df[("pointed_out_inspection_comment_count/input_data_count", phase)] = (
        df[("pointed_out_inspection_comment_count", phase)] / df[("input_data_count", phase)]
    )
    df[("rejected_count/task_count", phase)] = df[("rejected_count", phase)] / df[("task_count", phase)]


class AggregationBy(Enum):
    BY_INPUTS = "by_inputs"
    BY_TASKS = "by_tasks"


class AccountStatisticsValue(Enum):
    TASKS_COMPLETED = "tasks_completed"
    TASKS_REJECTED = "tasks_rejected"
    WORKTIME = "worktime"


class Table:
    """
    Databaseから取得した情報を元にPandas DataFrameを生成するクラス
    チェックポイントファイルがあること前提

    Attributes:
        label_dict: label_idとその名前の辞書
        inspection_phrases_dict: 定型指摘IDとそのコメント
        project_members_dict: account_idとプロジェクトメンバ情報

    """

    #############################################
    # Private
    #############################################

    _task_id_list: Optional[List[str]] = None
    _task_list: Optional[List[Task]] = None
    _inspections_dict: Optional[Dict[str, Dict[InputDataId, List[Inspection]]]] = None
    _input_data_dict: Optional[Dict[str, List[InputData]]] = None
    _task_histories_dict: Optional[Dict[str, List[TaskHistory]]] = None
    _annotations_dict: Optional[Dict[str, Dict[InputDataId, Dict[str, Any]]]] = None
    _worktime_statistics: Optional[List[WorktimeStatistics]] = None
    _account_statistics: Optional[List[ProjectAccountStatisticsHistory]] = None
    _labor_list: Optional[List[Dict[str, Any]]] = None

    def __init__(
        self,
        database: Database,
        ignored_task_id_list: Optional[List[str]] = None,
    ):
        self.annofab_service = database.annofab_service
        self.annofab_facade = AnnofabApiFacade(database.annofab_service)
        self.database = database
        self.ignored_task_id_list = ignored_task_id_list

        self.project_id = self.database.project_id
        self._update_annotaion_specs()
        self.project_title = self.annofab_service.api.get_project(self.project_id)[0]["title"]

    def _get_worktime_statistics(self) -> List[WorktimeStatistics]:
        """
        タスク作業時間集計を取得
        """
        if self._worktime_statistics is not None:
            return self._worktime_statistics
        else:
            tmp_worktime_statistics = self.annofab_service.wrapper.get_worktime_statistics(self.project_id)
            worktime_statistics: List[WorktimeStatistics] = [
                WorktimeStatistics.from_dict(e) for e in tmp_worktime_statistics
            ]
            self._worktime_statistics = worktime_statistics
            return worktime_statistics

    def _get_account_statistics(self) -> List[ProjectAccountStatisticsHistory]:
        """
        ユーザー別タスク集計を取得
        """
        if self._account_statistics is not None:
            return self._account_statistics
        else:
            content: List[Any] = self.annofab_service.wrapper.get_account_statistics(self.project_id)[0]
            account_statistics = [ProjectAccountStatisticsHistory.from_dict(e) for e in content]
            self._account_statistics = account_statistics
            return account_statistics

    def _get_task_list(self) -> List[Task]:
        """
        self.task_listを返す。Noneならば、self.task_list, self.task_id_listを設定する。
        """
        if self._task_list is not None:
            return self._task_list
        else:
            task_list = self.database.read_tasks_from_json()
            self._task_list = task_list
            return self._task_list

    def _get_inspections_dict(self) -> Dict[str, Dict[InputDataId, List[Inspection]]]:
        if self._inspections_dict is not None:
            return self._inspections_dict
        else:
            task_id_list = [t["task_id"] for t in self._get_task_list()]
            self._inspections_dict = self.database.read_inspections_from_json(task_id_list)
            return self._inspections_dict

    def _get_input_data_dict(self) -> Dict[str, List[InputData]]:
        if self._input_data_dict is not None:
            return self._input_data_dict
        else:
            self._input_data_dict = self.database.read_input_data_from_json(self._get_task_list())
            return self._input_data_dict

    def _get_task_histories_dict(self) -> Dict[str, List[TaskHistory]]:
        if self._task_histories_dict is not None:
            return self._task_histories_dict
        else:
            task_id_list = [t["task_id"] for t in self._get_task_list()]
            self._task_histories_dict = self.database.read_task_histories_from_json(task_id_list)
            return self._task_histories_dict

    def _get_annotations_dict(self) -> AnnotationDict:
        if self._annotations_dict is not None:
            return self._annotations_dict
        else:
            task_list = self._get_task_list()
            self._annotations_dict = self.database.read_annotation_summary(task_list, self._create_annotation_summary)
            return self._annotations_dict

    def _get_labor_list(self) -> List[Dict[str, Any]]:
        if self._labor_list is not None:
            return self._labor_list
        else:
            labor_list = self.database.read_labor_list_from_checkpoint()
            self._labor_list = labor_list
            return self._labor_list

    def _create_annotation_summary(self, annotation_list: List[Dict[str, Any]]) -> Dict[str, Any]:
        annotation_summary = {"total_count": len(annotation_list)}

        # labelごとのアノテーション数を算出
        for label_name in self.label_dict.values():
            annotation_count = 0
            key = f"label_{label_name}"
            annotation_count += len([e for e in annotation_list if e["label"] == label_name])
            annotation_summary[key] = annotation_count

        return annotation_summary

    @staticmethod
    def operator_is_changed_by_phase(task_history_list: List[TaskHistory], phase: TaskPhase) -> bool:
        """
        フェーズ内の作業者が途中で変わったかどうか

        Args:
            task_history_list: タスク履歴List
            phase: フェーズ

        Returns:
            Trueならばフェーズ内の作業者が途中で変わった
        """
        account_id_list = [
            e["account_id"] for e in task_history_list if TaskPhase(e["phase"]) == phase and e["account_id"] is not None
        ]
        return len(set(account_id_list)) >= 2

    @staticmethod
    def _inspection_condition(inspection_arg, exclude_reply: bool, only_error_corrected: bool):
        """
        対象の検査コメントかどうかの判定
        Args:
            inspection_arg:
            exclude_reply:
            only_error_corrected:

        Returns:

        """
        exclude_reply_flag = True
        if exclude_reply:
            exclude_reply_flag = inspection_arg["parent_inspection_id"] is None

        only_error_corrected_flag = True
        if only_error_corrected:
            only_error_corrected_flag = inspection_arg["status"] == InspectionStatus.ERROR_CORRECTED.value

        return exclude_reply_flag and only_error_corrected_flag

    def _get_user_id(self, account_id: Optional[str]) -> Optional[str]:
        """
        プロジェクトメンバのuser_idを取得する。プロジェクトメンバでなければ、account_idを返す。
        account_idがNoneならばNoneを返す。
        """
        if account_id is None:
            return None

        member = self.annofab_facade.get_project_member_from_account_id(self.project_id, account_id)
        if member is not None:
            return member["user_id"]
        else:
            return account_id

    def _get_username(self, account_id: Optional[str]) -> Optional[str]:
        """
        プロジェクトメンバのusernameを取得する。プロジェクトメンバでなければ、account_idを返す。
        account_idがNoneならばNoneを返す。
        """
        if account_id is None:
            return None

        member = self.annofab_facade.get_project_member_from_account_id(self.project_id, account_id)
        if member is not None:
            return member["username"]
        else:
            return account_id

    def _get_biography(self, account_id: Optional[str]) -> Optional[str]:
        """
        ユーザのbiographyを取得する。存在しないメンバであればNoneを返す。
        """
        if account_id is None:
            return None

        member = self.annofab_facade.get_project_member_from_account_id(self.project_id, account_id)
        if member is not None:
            return member["biography"]
        else:
            return None

    def _update_annotaion_specs(self):
        annotaion_specs = self.annofab_service.api.get_annotation_specs(self.project_id)[0]
        self.inspection_phrases_dict = self.get_inspection_phrases_dict(annotaion_specs["inspection_phrases"])
        self.label_dict = self.get_labels_dict(annotaion_specs["labels"])
        self.project_members_dict = self._get_project_members_dict()

    def _get_project_members_dict(self) -> Dict[str, Any]:
        project_members_dict = {}

        project_members = self.annofab_service.wrapper.get_all_project_members(self.project_id)
        for member in project_members:
            project_members_dict[member["account_id"]] = member
        return project_members_dict

    @staticmethod
    def get_labels_dict(labels) -> Dict[str, str]:
        """
        ラベル情報を設定する
        Returns:
            key: label_id, value: label_name(en)

        """
        label_dict = {}

        for e in labels:
            label_id = e["label_id"]
            messages_list = e["label_name"]["messages"]
            label_name = [m["message"] for m in messages_list if m["lang"] == "en-US"][0]

            label_dict[label_id] = label_name

        return label_dict

    @staticmethod
    def get_inspection_phrases_dict(inspection_phrases) -> Dict[str, str]:
        """
        定型指摘情報を取得
        Returns:
            定型指摘の辞書(key: id, value: 定型コメント(ja))

        """

        inspection_phrases_dict = {}

        for e in inspection_phrases:
            inspection_id = e["id"]
            messages_list = e["text"]["messages"]
            inspection_text = [m["message"] for m in messages_list if m["lang"] == "ja-JP"][0]

            inspection_phrases_dict[inspection_id] = inspection_text

        return inspection_phrases_dict

    def create_inspection_df(self, exclude_reply: bool = True, only_error_corrected: bool = True):
        """
        検査コメント一覧から、検査コメント用のdataframeを作成する。
        新たに追加した列は、"commenter_user_id","label_name"である.
        Args:
            exclude_reply: 返信コメント（「対応完了しました」など）を除く
            only_error_corrected: 「修正済」のコメントのみ。指摘はあったが、修正不要の場合は除外される。

        Returns:
            検査コメント一覧

        """

        master_task_list = self._get_task_list()
        inspections_dict = self._get_inspections_dict()

        all_inspection_list = []

        for task in master_task_list:
            task_id = task["task_id"]
            if task_id not in inspections_dict:
                continue

            input_data_dict = inspections_dict[task_id]
            for inspection_list in input_data_dict.values():

                # 検査コメントを絞り込む
                filtered_inspection_list = [
                    e for e in inspection_list if self._inspection_condition(e, exclude_reply, only_error_corrected)
                ]

                for inspection in filtered_inspection_list:
                    commenter_account_id = inspection["commenter_account_id"]
                    inspection["commenter_user_id"] = self._get_user_id(commenter_account_id)
                    inspection["commenter_username"] = self._get_username(commenter_account_id)

                    # inspection_phrases_dict に存在しないkeyを取得する可能性があるので、dict.getを用いる
                    inspection["phrases_name"] = [
                        self.inspection_phrases_dict.get(key) for key in inspection["phrases"]
                    ]

                    inspection["label_name"] = None
                    if inspection["label_id"] is not None:
                        inspection["label_name"] = self.label_dict.get(inspection["label_id"])

                all_inspection_list.extend(filtered_inspection_list)

        df = pandas.DataFrame(all_inspection_list)
        return df

    def _set_first_phase_from_task_history(
        self, task: Task, task_history: Optional[TaskHistory], column_prefix: str
    ) -> Task:
        """
        最初のフェーズに関する情報をTask情報に設定する。

        Args:
            task:
            task_history:
            column_prefix:

        Returns:

        """
        if task_history is None:
            task.update(
                {
                    f"{column_prefix}_account_id": None,
                    f"{column_prefix}_user_id": None,
                    f"{column_prefix}_username": None,
                    f"{column_prefix}_started_datetime": None,
                    f"{column_prefix}_worktime_hour": 0,
                }
            )
        else:
            account_id = task_history["account_id"]
            task.update(
                {
                    f"{column_prefix}_account_id": account_id,
                    f"{column_prefix}_user_id": self._get_user_id(account_id),
                    f"{column_prefix}_username": self._get_username(account_id),
                    f"{column_prefix}_started_datetime": task_history["started_datetime"],
                    f"{column_prefix}_worktime_hour": isoduration_to_hour(
                        task_history["accumulated_labor_time_milliseconds"]
                    ),
                }
            )

        return task

    @staticmethod
    def _get_first_operator_worktime(task_histories: List[TaskHistory], account_id: str) -> float:
        """
        対象ユーザが担当したフェーズの合計作業時間を取得する。

        Args:
            task_histories:
            account_id: 対象】ユーザのaccount_id

        Returns:
            作業時間

        """
        return sum(
            [
                isoduration_to_hour(e["accumulated_labor_time_milliseconds"])
                for e in task_histories
                if e["account_id"] == account_id
            ]
        )

    @staticmethod
    def _acceptance_is_skipped(task_histories: List[TaskHistory]) -> bool:
        task_history_index_list = get_task_history_index_skipped_acceptance(task_histories)
        if len(task_history_index_list) == 0:
            return False

        # スキップされた履歴より後に受入フェーズがなければ、受入がスキップされたタスクとみなす
        # ただし、スキップされた履歴より後で、「アノテーション一覧で修正された」受入フェーズがある場合（account_id is None）は、スキップされた受入とみなす。
        last_task_history_index = task_history_index_list[-1]
        return (
            more_itertools.first_true(
                task_histories[last_task_history_index + 1 :],
                pred=lambda e: e["phase"] == TaskPhase.ACCEPTANCE.value and e["account_id"] is not None,
            )
            is None
        )

    @staticmethod
    def _inspection_is_skipped(task_histories: List[TaskHistory]) -> bool:
        task_history_index_list = get_task_history_index_skipped_inspection(task_histories)
        if len(task_history_index_list) == 0:
            return False

        # スキップされた履歴より後に検査フェーズがなければ、検査がスキップされたタスクとみなす
        last_task_history_index = task_history_index_list[-1]
        return (
            more_itertools.first_true(
                task_histories[last_task_history_index + 1 :], pred=lambda e: e["phase"] == TaskPhase.INSPECTION.value
            )
            is None
        )

    def set_task_histories(self, task: Task, task_histories: List[TaskHistory]):
        """
        タスク履歴関係の情報を設定する
        """

        def diff_days(ended_key: str, started_key: str) -> Optional[float]:
            if task[ended_key] is not None and task[started_key] is not None:
                delta = dateutil.parser.parse(task[ended_key]) - dateutil.parser.parse(task[started_key])
                return delta.total_seconds() / 3600 / 24
            else:
                return None

        # タスク更新日維持と
        # if len(task_histories) > 0:
        #     # タスク情報とタスク履歴情報の整合性がとれているかを確認する
        #     delta = dateutil.parser.parse(task["updated_datetime"]) - dateutil.parser.parse(
        #         task_histories[-1]["ended_datetime"]
        #     )
        #     if abs(delta.total_seconds()) > 1:
        #         logger.warning(
        #             f"task_id={task['task_id']}のタスク情報とタスク履歴情報の整合性が取れていない可能性があります。"
        #             f"task.updated_datetime={task['updated_datetime']},"
        #             f"task_histories[-1].ended_datetime={task_histories[-1]['ended_datetime']}"
        #         )
        #
        annotation_histories = [
            e for e in task_histories if e["phase"] == TaskPhase.ANNOTATION.value and e["account_id"] is not None
        ]
        inspection_histories = [
            e for e in task_histories if e["phase"] == TaskPhase.INSPECTION.value and e["account_id"] is not None
        ]
        acceptance_histories = [
            e for e in task_histories if e["phase"] == TaskPhase.ACCEPTANCE.value and e["account_id"] is not None
        ]

        # 最初の教師付情報を設定する
        first_annotation_history = annotation_histories[0] if len(annotation_histories) > 0 else None
        self._set_first_phase_from_task_history(task, first_annotation_history, column_prefix="first_annotation")

        # 最初の検査情報を設定する
        first_inspection_history = inspection_histories[0] if len(inspection_histories) > 0 else None
        self._set_first_phase_from_task_history(task, first_inspection_history, column_prefix="first_inspection")

        # 最初の受入情報を設定する
        first_acceptance_history = acceptance_histories[0] if len(acceptance_histories) > 0 else None
        self._set_first_phase_from_task_history(task, first_acceptance_history, column_prefix="first_acceptance")

        task["annotation_worktime_hour"] = sum(
            [
                annofabcli.utils.isoduration_to_hour(e["accumulated_labor_time_milliseconds"])
                for e in annotation_histories
            ]
        )
        task["inspection_worktime_hour"] = sum(
            [
                annofabcli.utils.isoduration_to_hour(e["accumulated_labor_time_milliseconds"])
                for e in inspection_histories
            ]
        )
        task["acceptance_worktime_hour"] = sum(
            [
                annofabcli.utils.isoduration_to_hour(e["accumulated_labor_time_milliseconds"])
                for e in acceptance_histories
            ]
        )

        # 最初の教師者が担当した履歴の合計作業時間を取得する。
        # 担当者変更がなければ、"annotation_worktime_hour"と"first_annotation_worktime_hour"は同じ値
        task["first_annotator_worktime_hour"] = (
            self._get_first_operator_worktime(annotation_histories, first_annotation_history["account_id"])
            if first_annotation_history is not None
            else 0
        )

        task["first_inspector_worktime_hour"] = (
            self._get_first_operator_worktime(inspection_histories, first_inspection_history["account_id"])
            if first_inspection_history is not None
            else 0
        )
        task["first_acceptor_worktime_hour"] = (
            self._get_first_operator_worktime(acceptance_histories, first_acceptance_history["account_id"])
            if first_acceptance_history is not None
            else 0
        )

        task["sum_worktime_hour"] = sum(
            [annofabcli.utils.isoduration_to_hour(e["accumulated_labor_time_milliseconds"]) for e in task_histories]
        )

        task["number_of_rejections_by_inspection"] = get_number_of_rejections(
            task["histories_by_phase"], TaskPhase.INSPECTION
        )
        task["number_of_rejections_by_acceptance"] = get_number_of_rejections(
            task["histories_by_phase"], TaskPhase.ACCEPTANCE
        )
        # APIで取得した 'number_of_rejections' は非推奨で、number_of_rejections_by_inspection/acceptanceと矛盾する場合があるので、書き換える
        task["number_of_rejections"] = (
            task["number_of_rejections_by_inspection"] + task["number_of_rejections_by_acceptance"]
        )

        # 受入完了日時を設定
        if task["phase"] == TaskPhase.ACCEPTANCE.value and task["status"] == TaskStatus.COMPLETE.value:
            if len(acceptance_histories) > 0:
                task["task_completed_datetime"] = acceptance_histories[-1]["ended_datetime"]
            else:
                logger.warning(f"task_id={task['task_id']}のタスク履歴が間違っている可能性があります。")
                task["task_completed_datetime"] = None
        else:
            task["task_completed_datetime"] = None

        task["diff_days_to_first_inspection_started"] = diff_days(
            "first_inspection_started_datetime", "first_annotation_started_datetime"
        )
        task["diff_days_to_first_acceptance_started"] = diff_days(
            "first_acceptance_started_datetime", "first_annotation_started_datetime"
        )

        task["diff_days_to_task_completed"] = diff_days("task_completed_datetime", "first_annotation_started_datetime")

        # 抜取検査/抜取受入で、検査/受入がスキップされたか否か
        task["acceptance_is_skipped"] = self._acceptance_is_skipped(task_histories)
        task["inspection_is_skipped"] = self._inspection_is_skipped(task_histories)
        task["annotator_is_changed"] = self.operator_is_changed_by_phase(task_histories, TaskPhase.ANNOTATION)
        task["inspector_is_changed"] = self.operator_is_changed_by_phase(task_histories, TaskPhase.INSPECTION)
        task["acceptor_is_changed"] = self.operator_is_changed_by_phase(task_histories, TaskPhase.ACCEPTANCE)

        return task

    @staticmethod
    def _get_task_from_task_id(task_list: List[Task], task_id: str) -> Optional[Task]:
        return more_itertools.first_true(task_list, pred=lambda e: e["task_id"] == task_id)

    def create_task_history_df(self) -> pandas.DataFrame:
        """
        タスク履歴の一覧のDataFrameを出力する。

        Returns:

        """
        task_histories_dict = self._get_task_histories_dict()
        task_list = self._get_task_list()

        all_task_history_list = []
        for task_id, task_history_list in task_histories_dict.items():
            task = self._get_task_from_task_id(task_list, task_id)
            if task is None:
                continue

            for history in task_history_list:
                account_id = history["account_id"]
                history["user_id"] = self._get_user_id(account_id)
                history["username"] = self._get_username(account_id)
                history["biography"] = self._get_biography(account_id)
                history["worktime_hour"] = annofabcli.utils.isoduration_to_hour(
                    history["accumulated_labor_time_milliseconds"]
                )
                # task statusがあると分析しやすいので追加する
                history["task_status"] = task["status"]
                all_task_history_list.append(history)

        df = pandas.DataFrame(all_task_history_list)
        if len(df) == 0:
            logger.warning("タスク履歴の件数が0件です。")
            return pandas.DataFrame()
        else:
            return df

    def create_task_df(self) -> pandas.DataFrame:
        """
        タスク一覧からdataframeを作成する。
        新たに追加した列は、user_id, annotation_count, inspection_count,
            first_annotation_user_id, first_annotation_started_datetime,
            annotation_worktime_hour, inspection_worktime_hour, acceptance_worktime_hour, sum_worktime_hour
        """

        def set_input_data_info(arg_task):
            input_data_list = input_data_dict.get(arg_task["task_id"])
            input_duration_list = [e["input_duration"] for e in input_data_list]
            if None in input_duration_list:
                # 動画プロジェクトや動画時間が設定されていない場合
                arg_task["input_duration_seconds"] = None
            else:
                arg_task["input_duration_seconds"] = sum(input_duration_list)

        def set_annotation_info(arg_task):
            total_annotation_count = 0
            input_data_id_list = arg_task["input_data_id_list"]
            input_data_dict = annotations_dict.get(arg_task["task_id"], None)
            if input_data_dict is not None:
                for input_data_id in input_data_id_list:
                    total_annotation_count += input_data_dict[input_data_id]["total_count"]
            arg_task["annotation_count"] = total_annotation_count

        def set_inspection_info(arg_task):
            # 指摘枚数
            inspection_count = 0
            # 指摘された画像枚数
            input_data_count_of_inspection = 0

            input_data_dict = inspections_dict.get(arg_task["task_id"])
            if input_data_dict is not None:
                for inspection_list in input_data_dict.values():
                    # 検査コメントを絞り込む
                    filtered_inspection_list = [
                        e
                        for e in inspection_list
                        if self._inspection_condition(e, exclude_reply=True, only_error_corrected=True)
                    ]
                    inspection_count += len(filtered_inspection_list)
                    if len(filtered_inspection_list) > 0:
                        input_data_count_of_inspection += 1

            arg_task["inspection_count"] = inspection_count
            arg_task["input_data_count_of_inspection"] = input_data_count_of_inspection

        tasks = self._get_task_list()
        task_histories_dict = self._get_task_histories_dict()
        inspections_dict = self._get_inspections_dict()
        annotations_dict = self._get_annotations_dict()
        input_data_dict = self._get_input_data_dict()

        for task in tasks:
            task_id = task["task_id"]
            task_histories = task_histories_dict.get(task_id, [])

            account_id = task["account_id"]
            task["user_id"] = self._get_user_id(account_id)
            task["username"] = self._get_username(account_id)
            task["input_data_count"] = len(task["input_data_id_list"])

            self.set_task_histories(task, task_histories)
            set_annotation_info(task)
            set_inspection_info(task)
            set_input_data_info(task)

        df = pandas.DataFrame(tasks)
        if len(df) > 0:
            # dictが含まれたDataFrameをbokehでグラフ化するとErrorが発生するので、dictを含む列を削除する
            # https://github.com/bokeh/bokeh/issues/9620
            df = df.drop(["histories_by_phase"], axis=1)
            return df
        else:
            logger.warning(f"タスク一覧が0件です。")
            return pandas.DataFrame()

    def create_task_for_annotation_df(self):
        """
        アノテーションラベルごとのアノテーション数を表示したタスク一覧を、取得する。
        アノテーションラベルのアノテーション数を格納した列名は、アノテーションラベルの日本語名になる。
        """

        tasks = self._get_task_list()
        annotations_dict = self._get_annotations_dict()

        task_list = []
        for task in tasks:
            task_id = task["task_id"]
            new_task = {}
            for key in ["task_id", "phase", "status"]:
                new_task[key] = task[key]

            new_task["input_data_count"] = len(task["input_data_id_list"])
            input_data_id_list = task["input_data_id_list"]

            input_data_dict = annotations_dict.get(task_id, None)
            if input_data_dict is not None:
                for label_name in self.label_dict.values():
                    annotation_count = 0
                    for input_data_id in input_data_id_list:
                        annotation_summary = input_data_dict[input_data_id]
                        annotation_count += annotation_summary[f"label_{label_name}"]

                    new_task[f"label_{label_name}"] = annotation_count
            else:
                for label_name in self.label_dict.values():
                    new_task[f"label_{label_name}"] = 0

            task_list.append(new_task)

        df = pandas.DataFrame(task_list)
        return df

    @staticmethod
    def create_dataframe_by_date_user_for_annotation(task_df: pandas.DataFrame) -> pandas.DataFrame:
        """
        日毎、ユーザごとの情報を出力する。

        Args:
            task_df:

        Returns:

        """
        new_df = task_df
        new_df["first_annotation_started_date"] = new_df["first_annotation_started_datetime"].map(
            lambda e: datetime_to_date(e) if e is not None and isinstance(e, str) else None
        )
        new_df["task_count"] = 1  # 集計用

        # first_annotation_user_id と first_annotation_usernameの両方を指定している理由：
        # first_annotation_username を取得するため
        group_obj = new_df.groupby(
            ["first_annotation_started_date", "first_annotation_user_id", "first_annotation_username"],
            as_index=False,
        )
        sum_df = group_obj[
            [
                "first_annotation_worktime_hour",
                "annotation_worktime_hour",
                "inspection_worktime_hour",
                "acceptance_worktime_hour",
                "sum_worktime_hour",
                "task_count",
                "input_data_count",
                "annotation_count",
                "inspection_count",
            ]
        ].sum()

        sum_df["first_annotation_worktime_minute/annotation_count"] = (
            sum_df["first_annotation_worktime_hour"] * 60 / sum_df["annotation_count"]
        )
        sum_df["annotation_worktime_minute/annotation_count"] = (
            sum_df["annotation_worktime_hour"] * 60 / sum_df["annotation_count"]
        )
        sum_df["inspection_worktime_minute/annotation_count"] = (
            sum_df["inspection_worktime_hour"] * 60 / sum_df["annotation_count"]
        )
        sum_df["acceptance_worktime_minute/annotation_count"] = (
            sum_df["acceptance_worktime_hour"] * 60 / sum_df["annotation_count"]
        )
        sum_df["inspection_count/annotation_count"] = sum_df["inspection_count"] / sum_df["annotation_count"]

        sum_df["first_annotation_worktime_minute/input_data_count"] = (
            sum_df["first_annotation_worktime_hour"] * 60 / sum_df["input_data_count"]
        )
        sum_df["annotation_worktime_minute/input_data_count"] = (
            sum_df["annotation_worktime_hour"] * 60 / sum_df["input_data_count"]
        )
        sum_df["inspection_worktime_minute/input_data_count"] = (
            sum_df["inspection_worktime_hour"] * 60 / sum_df["input_data_count"]
        )
        sum_df["acceptance_worktime_minute/input_data_count"] = (
            sum_df["acceptance_worktime_hour"] * 60 / sum_df["input_data_count"]
        )
        sum_df["inspection_count/input_data_count"] = sum_df["inspection_count"] / sum_df["annotation_count"]

        return sum_df

    @staticmethod
    def create_dataframe_by_date_user_for_inspection(task_df: pandas.DataFrame) -> pandas.DataFrame:
        """
        検査者用に日毎、ユーザごとの情報を出力する。

        Args:
            task_df:

        Returns:

        """
        new_df = task_df
        new_df["first_inspection_started_date"] = new_df["first_inspection_started_datetime"].map(
            lambda e: datetime_to_date(e) if e is not None and isinstance(e, str) else None
        )
        new_df["task_count"] = 1  # 集計用

        # first_inspection_user_id と first_inspection_username の両方を指定している理由：
        # first_inspection_username を取得するため
        group_obj = new_df.groupby(
            ["first_inspection_started_date", "first_inspection_user_id", "first_inspection_username"],
            as_index=False,
        )
        sum_df = group_obj[
            [
                "first_inspection_worktime_hour",
                "annotation_worktime_hour",
                "inspection_worktime_hour",
                "acceptance_worktime_hour",
                "sum_worktime_hour",
                "task_count",
                "input_data_count",
                "annotation_count",
                "inspection_count",
            ]
        ].sum()

        sum_df["first_inspection_worktime_minute/annotation_count"] = (
            sum_df["first_inspection_worktime_hour"] * 60 / sum_df["annotation_count"]
        )
        sum_df["inspection_worktime_minute/annotation_count"] = (
            sum_df["inspection_worktime_hour"] * 60 / sum_df["annotation_count"]
        )

        sum_df["first_inspection_worktime_minute/input_data_count"] = (
            sum_df["first_inspection_worktime_hour"] * 60 / sum_df["input_data_count"]
        )
        sum_df["inspection_worktime_minute/input_data_count"] = (
            sum_df["inspection_worktime_hour"] * 60 / sum_df["input_data_count"]
        )

        return sum_df

    @staticmethod
    def create_dataframe_by_date_user_for_acceptance(task_df: pandas.DataFrame) -> pandas.DataFrame:
        """
        受入用に日毎、ユーザごとの情報を出力する。

        Args:
            task_df:

        Returns:

        """
        new_df = task_df
        new_df["first_acceptance_started_date"] = new_df["first_acceptance_started_datetime"].map(
            lambda e: datetime_to_date(e) if e is not None and isinstance(e, str) else None
        )
        new_df["task_count"] = 1  # 集計用

        # first_acceptance_user_id と first_acceptance_username の両方を指定している理由：
        # first_acceptance_username を取得するため
        group_obj = new_df.groupby(
            ["first_acceptance_started_date", "first_acceptance_user_id", "first_acceptance_username"],
            as_index=False,
        )
        sum_df = group_obj[
            [
                "first_acceptance_worktime_hour",
                "annotation_worktime_hour",
                "inspection_worktime_hour",
                "acceptance_worktime_hour",
                "sum_worktime_hour",
                "task_count",
                "input_data_count",
                "annotation_count",
                "inspection_count",
            ]
        ].sum()

        sum_df["first_acceptance_worktime_minute/annotation_count"] = (
            sum_df["first_acceptance_worktime_hour"] * 60 / sum_df["annotation_count"]
        )
        sum_df["acceptance_worktime_minute/annotation_count"] = (
            sum_df["acceptance_worktime_hour"] * 60 / sum_df["annotation_count"]
        )

        sum_df["first_acceptance_worktime_minute/input_data_count"] = (
            sum_df["first_acceptance_worktime_hour"] * 60 / sum_df["input_data_count"]
        )
        sum_df["acceptance_worktime_minute/input_data_count"] = (
            sum_df["acceptance_worktime_hour"] * 60 / sum_df["input_data_count"]
        )

        return sum_df

    def create_member_df(self, task_df: pandas.DataFrame) -> pandas.DataFrame:
        """
        プロジェクトメンバ一覧の情報
        """

        task_histories_dict = self._get_task_histories_dict()

        member_dict = {}
        for account_id, member in self.project_members_dict.items():
            new_member = copy.deepcopy(member)
            new_member.update(
                {
                    # 初回のアノテーションに関わった個数（タスクの教師付担当者は変更されない前提）
                    "task_count_of_first_annotation": 0,
                    "input_data_count_of_first_annotation": 0,
                    "annotation_count_of_first_annotation": 0,
                    "inspection_count_of_first_annotation": 0,
                    # 関わった作業時間
                    "annotation_worktime_hour": 0,
                    "inspection_worktime_hour": 0,
                    "acceptance_worktime_hour": 0,
                }
            )
            member_dict[account_id] = new_member

        for _, row_task in task_df.iterrows():
            task_id = row_task["task_id"]

            task_histories = task_histories_dict.get(task_id, [])

            # 初回のアノテーションに関わった個数
            if row_task["first_annotation_account_id"] is not None:
                first_annotation_account_id = row_task["first_annotation_account_id"]
                if first_annotation_account_id in member_dict:
                    member = member_dict[first_annotation_account_id]
                    member["task_count_of_first_annotation"] += 1
                    member["input_data_count_of_first_annotation"] += row_task["input_data_count"]
                    member["annotation_count_of_first_annotation"] += row_task["annotation_count"]
                    member["inspection_count_of_first_annotation"] += row_task["inspection_count"]

            # 関わった作業時間を記載
            for history in task_histories:
                account_id = history["account_id"]
                if account_id is None:
                    continue

                phase = history["phase"]
                worktime_hour = annofabcli.utils.isoduration_to_hour(history["accumulated_labor_time_milliseconds"])
                if history["account_id"] in member_dict:
                    member = member_dict[history["account_id"]]
                    member[f"{phase}_worktime_hour"] += worktime_hour

        member_list = []
        for account_id, member in member_dict.items():
            member_list.append(member)

        df = pandas.DataFrame(member_list)
        return df

    def create_account_statistics_df(self):
        """
        メンバごと、日ごとの作業時間を、統計情報APIから取得する。
        """

        account_statistics = self.database.read_account_statistics_from_checkpoint()

        all_histories = []
        for account_info in account_statistics:

            account_id = account_info["account_id"]
            histories = account_info["histories"]

            member_info = self.project_members_dict.get(account_id)

            for history in histories:
                history["worktime_hour"] = annofabcli.utils.isoduration_to_hour(history["worktime"])
                history["account_id"] = account_id
                history["user_id"] = self._get_user_id(account_id)
                history["username"] = self._get_username(account_id)
                history["biography"] = self._get_biography(account_id)

                if member_info is not None:
                    history["member_role"] = member_info["member_role"]

            all_histories.extend(histories)

        df = pandas.DataFrame(all_histories)
        return df

    @staticmethod
    def create_cumulative_df_by_user(account_statistics_df: pandas.DataFrame) -> pandas.DataFrame:
        """
        アカウントごとの作業時間に対して、累積作業時間を追加する。
        Args:
            account_statistics_df: アカウントごとの作業時間用DataFrame

        Returns:

        """
        # 教師付の開始時刻でソートして、indexを更新する
        df = account_statistics_df.sort_values(["account_id", "date"]).reset_index(drop=True)
        groupby_obj = df.groupby("account_id")
        df["cumulative_worktime_hour"] = groupby_obj["worktime_hour"].cumsum()
        return df

    @staticmethod
    def create_cumulative_df_overall(task_df: pandas.DataFrame) -> pandas.DataFrame:
        """
        1回目の教師付開始日でソートして、累積値を算出する。

        Args:
            task_df: タスク一覧のDataFrame. 列が追加される
        """
        # 教師付の開始時刻でソートして、indexを更新する
        if len(task_df) == 0:
            logger.warning(f"タスク一覧が0件です。")
            return pandas.DataFrame()

        df = task_df.sort_values(["first_annotation_started_datetime"]).reset_index(drop=True)
        # タスクの累計数を取得するために設定する
        df["task_count"] = 1

        # 作業時間の累積値
        df["cumulative_annotation_worktime_hour"] = df["annotation_worktime_hour"].cumsum()
        df["cumulative_inspection_worktime_hour"] = df["inspection_worktime_hour"].cumsum()
        df["cumulative_acceptance_worktime_hour"] = df["acceptance_worktime_hour"].cumsum()
        df["cumulative_sum_worktime_hour"] = df["sum_worktime_hour"].cumsum()

        # タスク完了数、差し戻し数など
        df["cumulative_inspection_count"] = df["inspection_count"].cumsum()
        df["cumulative_annotation_count"] = df["annotation_count"].cumsum()
        df["cumulative_input_data_count"] = df["input_data_count"].cumsum()
        df["cumulative_task_count"] = df["task_count"].cumsum()
        df["cumulative_number_of_rejections"] = df["number_of_rejections"].cumsum()
        df["cumulative_number_of_rejections_by_inspection"] = df["number_of_rejections_by_inspection"].cumsum()
        df["cumulative_number_of_rejections_by_acceptance"] = df["number_of_rejections_by_acceptance"].cumsum()

        # 元に戻す
        df = df.drop(["task_count"], axis=1)
        return df

    @staticmethod
    def create_cumulative_df_by_first_annotator(task_df: pandas.DataFrame) -> pandas.DataFrame:
        """
        最初のアノテーション作業の開始時刻の順にソートして、教師付者に関する累計値を算出する
        Args:
            task_df: タスク一覧のDataFrame. 列が追加される
        """
        # 教師付の開始時刻でソートして、indexを更新する
        df = task_df.sort_values(["first_annotation_user_id", "first_annotation_started_datetime"]).reset_index(
            drop=True
        )
        # タスクの累計数を取得するために設定する
        df["task_count"] = 1
        # 教師付の作業者でgroupby
        groupby_obj = df.groupby("first_annotation_user_id")

        # 作業時間の累積値
        df["cumulative_annotation_worktime_hour"] = groupby_obj["annotation_worktime_hour"].cumsum()
        df["cumulative_acceptance_worktime_hour"] = groupby_obj["acceptance_worktime_hour"].cumsum()
        df["cumulative_inspection_worktime_hour"] = groupby_obj["inspection_worktime_hour"].cumsum()
        df["cumulative_first_annotation_worktime_hour"] = groupby_obj["first_annotation_worktime_hour"].cumsum()

        # タスク完了数、差し戻し数など
        df["cumulative_inspection_count"] = groupby_obj["inspection_count"].cumsum()
        df["cumulative_annotation_count"] = groupby_obj["annotation_count"].cumsum()
        df["cumulative_input_data_count"] = groupby_obj["input_data_count"].cumsum()
        df["cumulative_task_count"] = groupby_obj["task_count"].cumsum()
        df["cumulative_number_of_rejections"] = groupby_obj["number_of_rejections"].cumsum()
        df["cumulative_number_of_rejections_by_inspection"] = groupby_obj["number_of_rejections_by_inspection"].cumsum()
        df["cumulative_number_of_rejections_by_acceptance"] = groupby_obj["number_of_rejections_by_acceptance"].cumsum()

        # 元に戻す
        df = df.drop(["task_count"], axis=1)
        return df

    @staticmethod
    def create_gradient_df(task_df: pandas.DataFrame) -> pandas.DataFrame:
        """
        生産量あたりの指標を算出する

        """
        task_df["annotation_worktime_hour/annotation_count"] = (
            task_df["annotation_worktime_hour"] / task_df["annotation_count"]
        )
        task_df["inspection_count/annotation_count"] = task_df["inspection_count"] / task_df["annotation_count"]
        task_df["number_of_rejections/annotation_count"] = task_df["number_of_rejections"] / task_df["annotation_count"]

        task_df["inspection_worktime_hour/annotation_count"] = (
            task_df["inspection_worktime_hour"] / task_df["annotation_count"]
        )
        task_df["acceptance_worktime_hour/annotation_count"] = (
            task_df["acceptance_worktime_hour"] / task_df["annotation_count"]
        )
        return task_df

    @staticmethod
    def create_cumulative_df_by_first_inspector(task_df: pandas.DataFrame) -> pandas.DataFrame:
        """
        最初の検査作業の開始時刻の順にソートして、累計値を算出する
        Args:
            task_df: タスク一覧のDataFrame. 列が追加される
        """

        df = task_df.sort_values(["first_inspection_user_id", "first_inspection_started_datetime"]).reset_index(
            drop=True
        )
        # タスクの累計数を取得するために設定する
        df["task_count"] = 1
        # 教師付の作業者でgroupby
        groupby_obj = df.groupby("first_inspection_user_id")

        # 作業時間の累積値
        df["cumulative_annotation_worktime_hour"] = groupby_obj["annotation_worktime_hour"].cumsum()
        df["cumulative_acceptance_worktime_hour"] = groupby_obj["acceptance_worktime_hour"].cumsum()
        df["cumulative_inspection_worktime_hour"] = groupby_obj["inspection_worktime_hour"].cumsum()
        df["cumulative_first_inspection_worktime_hour"] = groupby_obj["first_inspection_worktime_hour"].cumsum()

        # タスク完了数、差し戻し数など
        df["cumulative_inspection_count"] = groupby_obj["inspection_count"].cumsum()
        df["cumulative_annotation_count"] = groupby_obj["annotation_count"].cumsum()
        df["cumulative_input_data_count"] = groupby_obj["input_data_count"].cumsum()
        df["cumulative_task_count"] = groupby_obj["task_count"].cumsum()
        df["cumulative_number_of_rejections"] = groupby_obj["number_of_rejections"].cumsum()
        df["cumulative_number_of_rejections_by_inspection"] = groupby_obj["number_of_rejections_by_inspection"].cumsum()
        df["cumulative_number_of_rejections_by_acceptance"] = groupby_obj["number_of_rejections_by_acceptance"].cumsum()

        # 元に戻す
        df = df.drop(["task_count"], axis=1)
        return df

    @staticmethod
    def create_cumulative_df_by_first_acceptor(task_df: pandas.DataFrame) -> pandas.DataFrame:
        """
        最初の受入作業の開始時刻の順にソートして、累計値を算出する
        Args:
            task_df: タスク一覧のDataFrame. 列が追加される
        """

        df = task_df.sort_values(["first_acceptance_user_id", "first_acceptance_started_datetime"]).reset_index(
            drop=True
        )
        # タスクの累計数を取得するために設定する
        df["task_count"] = 1
        groupby_obj = df.groupby("first_acceptance_user_id")

        # 作業時間の累積値
        df["cumulative_acceptance_worktime_hour"] = groupby_obj["acceptance_worktime_hour"].cumsum()
        df["cumulative_first_acceptance_worktime_hour"] = groupby_obj["first_acceptance_worktime_hour"].cumsum()

        # タスク完了数、差し戻し数など
        df["cumulative_inspection_count"] = groupby_obj["inspection_count"].cumsum()
        df["cumulative_annotation_count"] = groupby_obj["annotation_count"].cumsum()
        df["cumulative_input_data_count"] = groupby_obj["input_data_count"].cumsum()
        df["cumulative_task_count"] = groupby_obj["task_count"].cumsum()
        df["cumulative_number_of_rejections"] = groupby_obj["number_of_rejections"].cumsum()
        df["cumulative_number_of_rejections_by_inspection"] = groupby_obj["number_of_rejections_by_inspection"].cumsum()
        df["cumulative_number_of_rejections_by_acceptance"] = groupby_obj["number_of_rejections_by_acceptance"].cumsum()

        # 元に戻す
        df = df.drop(["task_count"], axis=1)
        return df

    def create_worktime_per_image_df(self, aggregation_by: AggregationBy, phase: TaskPhase) -> pandas.DataFrame:
        """
        画像１枚あたり/タスク１個あたりの作業時間を算出する。
        行方向に日付, 列方向にメンバを並べる

        Args:
            aggregation_by: 集計単位（画像1枚あたり or タスク１個あたり）
            phase: 対象のフェーズ

        Returns:
            DataFrame
        """

        worktime_statistics = self._get_worktime_statistics()

        worktime_info_list = []
        for elm in worktime_statistics:
            worktime_info: Dict[str, Any] = {"date": elm.date}
            for account_info in elm.accounts:
                stat_list = getattr(account_info, aggregation_by.value)
                stat_item: Optional[WorktimeStatisticsItem] = first_true(stat_list, pred=lambda e: e.phase == phase)
                if stat_item is not None:
                    worktime_info[account_info.account_id] = isoduration_to_hour(stat_item.average)
                else:
                    worktime_info[account_info.account_id] = 0

            worktime_info_list.append(worktime_info)

        df = pandas.DataFrame(worktime_info_list)
        # acount_idをusernameに変更する
        columns = {
            col: self._get_username(col) for col in df.columns if col != "date"  # pylint: disable=not-an-iterable
        }
        return df.rename(columns=columns).fillna(0)

    @staticmethod
    def create_annotation_count_ratio_df(
        task_history_df: pandas.DataFrame, task_df: pandas.DataFrame
    ) -> pandas.DataFrame:
        """
        task_id, phase, (phase_index), user_idの作業時間比から、アノテーション数などの生産量を求める

        Args:

        Returns:

        """
        annotation_count_dict = {
            row["task_id"]: {
                "annotation_count": row["annotation_count"],
                "input_data_count": row["input_data_count"],
                "inspection_comment_count": row["inspection_count"],
                "rejected_count": row["number_of_rejections"],
            }
            for _, row in task_df.iterrows()
        }

        def get_inspection_comment_count(row) -> float:
            task_id = row.name[0]
            inspection_comment_count = annotation_count_dict[task_id]["inspection_comment_count"]
            return row["worktime_ratio_by_task"] * inspection_comment_count

        def get_rejected_count(row) -> float:
            task_id = row.name[0]
            rejected_count = annotation_count_dict[task_id]["rejected_count"]
            return row["worktime_ratio_by_task"] * rejected_count

        def get_annotation_count(row) -> float:
            task_id = row.name[0]
            annotation_count = annotation_count_dict[task_id]["annotation_count"]
            return row["worktime_ratio_by_task"] * annotation_count

        def get_input_data_count(row) -> float:
            task_id = row.name[0]
            annotation_count = annotation_count_dict[task_id]["input_data_count"]
            return row["worktime_ratio_by_task"] * annotation_count

        if len(task_df) == 0:
            logger.warning("タスク一覧が0件です。")
            return pandas.DataFrame()

        group_obj = task_history_df.groupby(["task_id", "phase", "phase_stage", "user_id"]).agg(
            {"worktime_hour": "sum"}
        )
        group_obj["worktime_ratio_by_task"] = group_obj.groupby(level=["task_id", "phase", "phase_stage"]).apply(
            lambda e: e / float(e.sum())
        )
        group_obj["annotation_count"] = group_obj.apply(get_annotation_count, axis="columns")
        group_obj["input_data_count"] = group_obj.apply(get_input_data_count, axis="columns")
        group_obj["pointed_out_inspection_comment_count"] = group_obj.apply(
            get_inspection_comment_count, axis="columns"
        )
        group_obj["rejected_count"] = group_obj.apply(get_rejected_count, axis="columns")
        new_df = group_obj.reset_index()
        new_df["pointed_out_inspection_comment_count"] = new_df["pointed_out_inspection_comment_count"] * new_df[
            "phase"
        ].apply(lambda e: 1 if e == TaskPhase.ANNOTATION.value else 0)
        new_df["rejected_count"] = new_df["rejected_count"] * new_df["phase"].apply(
            lambda e: 1 if e == TaskPhase.ANNOTATION.value else 0
        )
        return new_df

    @staticmethod
    def _get_phase_list(columns: List[str]) -> List[str]:
        phase_list = [TaskPhase.ANNOTATION.value, TaskPhase.INSPECTION.value, TaskPhase.ACCEPTANCE.value]
        if TaskPhase.INSPECTION.value not in columns:
            phase_list.remove(TaskPhase.INSPECTION.value)

        if TaskPhase.ACCEPTANCE.value not in columns:
            phase_list.remove(TaskPhase.ACCEPTANCE.value)

        return phase_list

    @staticmethod
    def merge_productivity_per_user_from_aw_time(df1: pandas.DataFrame, df2: pandas.DataFrame) -> pandas.DataFrame:
        """
        ユーザごとの生産性・品質情報が格納されたDataFrameを結合する
        Args:
            df1:
            df2:

        Returns:
            マージ済のユーザごとの生産性・品質情報
        """

        def max_last_working_date(date1, date2):
            if not isinstance(date1, str) and numpy.isnan(date1):
                date1 = ""
            if not isinstance(date2, str) and numpy.isnan(date2):
                date2 = ""
            max_date = max(date1, date2)
            if max_date == "":
                return numpy.nan
            else:
                return max_date

        def merge_row(row1: pandas.Series, row2: pandas.Series) -> pandas.Series:
            string_column_list = ["username", "biography", "last_working_date"]
            sum_row = row1.drop(labels=string_column_list, level=0).fillna(0) + row2.drop(
                labels=string_column_list, level=0
            ).fillna(0)
            sum_row.loc["username", ""] = row1.loc["username", ""]
            sum_row.loc["biography", ""] = row1.loc["biography", ""]
            sum_row.loc["last_working_date", ""] = max_last_working_date(
                row1.loc["last_working_date", ""], row2.loc["last_working_date", ""]
            )
            return sum_row

        user_id_set = set(df1["user_id"]) | set(df2["user_id"])
        sum_df = df1.set_index("user_id").copy()
        added_df = df2.set_index("user_id")

        for user_id in user_id_set:
            if user_id not in added_df.index:
                continue
            if user_id in sum_df.index:
                sum_df.loc[user_id] = merge_row(sum_df.loc[user_id], added_df.loc[user_id])
            else:
                sum_df.loc[user_id] = added_df.loc[user_id]

        phase_list = Table._get_phase_list(list(sum_df["monitored_worktime_hour"].columns))
        _add_ratio_column_for_productivity_per_user(sum_df, phase_list=phase_list)
        return sum_df.reset_index().sort_values(["user_id"])

    @staticmethod
    def create_productivity_per_user_from_aw_time(
        df_task_history: pandas.DataFrame, df_labor: pandas.DataFrame, df_worktime_ratio: pandas.DataFrame
    ) -> pandas.DataFrame:
        """
        AnnoWorkの実績時間から、作業者ごとに生産性を算出する。

        Returns:

        """
        df_agg_task_history = df_task_history.pivot_table(
            values="worktime_hour", columns="phase", index="user_id", aggfunc=numpy.sum
        ).fillna(0)

        if len(df_labor) > 0:
            df_agg_labor = df_labor.pivot_table(values="worktime_result_hour", index="user_id", aggfunc=numpy.sum)
            df_tmp = df_labor[df_labor["worktime_result_hour"] > 0].pivot_table(
                values="date", index="user_id", aggfunc=numpy.max
            )
            if len(df_tmp) > 0:
                df_agg_labor["last_working_date"] = df_tmp
            else:
                df_agg_labor["last_working_date"] = numpy.nan
            df = df_agg_task_history.join(df_agg_labor)
        else:
            df = df_agg_task_history
            df["worktime_result_hour"] = 0
            df["last_working_date"] = None

        phase_list = Table._get_phase_list(list(df.columns))

        df = df[["worktime_result_hour", "last_working_date"] + phase_list].copy()
        df.columns = pandas.MultiIndex.from_tuples(
            [("actual_worktime_hour", "sum"), ("last_working_date", "")]
            + [("monitored_worktime_hour", phase) for phase in phase_list]
        )

        df[("monitored_worktime_hour", "sum")] = df[[("monitored_worktime_hour", phase) for phase in phase_list]].sum(
            axis=1
        )

        df_agg_production = df_worktime_ratio.pivot_table(
            values=[
                "worktime_ratio_by_task",
                "input_data_count",
                "annotation_count",
                "pointed_out_inspection_comment_count",
                "rejected_count",
            ],
            columns="phase",
            index="user_id",
            aggfunc=numpy.sum,
        ).fillna(0)

        df_agg_production.rename(columns={"worktime_ratio_by_task": "task_count"}, inplace=True)
        df = df.join(df_agg_production)

        # 比例関係の列を計算して追加する
        _add_ratio_column_for_productivity_per_user(df, phase_list=phase_list)

        # 不要な列を削除する
        tmp_phase_list = copy.deepcopy(phase_list)
        tmp_phase_list.remove(TaskPhase.ANNOTATION.value)

        dropped_column = [("pointed_out_inspection_comment_count", phase) for phase in tmp_phase_list] + [
            ("rejected_count", phase) for phase in tmp_phase_list
        ]
        df = df.drop(dropped_column, axis=1)

        # ユーザ情報を取得
        df_user = df_task_history.groupby("user_id").first()[["username", "biography"]]
        df_user.columns = pandas.MultiIndex.from_tuples([("username", ""), ("biography", "")])
        df = df.join(df_user)
        df[("user_id", "")] = df.index

        return df

    def create_labor_df(self) -> pandas.DataFrame:
        """
        労務管理 DataFrameを生成する。情報を出力する。

        """

        def add_user_info(d: Dict[str, Any]) -> Dict[str, Any]:
            account_id = d["account_id"]
            d["user_id"] = self._get_user_id(account_id)
            d["username"] = self._get_username(account_id)
            d["biography"] = self._get_biography(account_id)
            return d

        labor_list = self._get_labor_list()
        return pandas.DataFrame([add_user_info(e) for e in labor_list])

    @staticmethod
    def _create_date_df(date_index1: pandas.Index, date_index2: pandas.Index) -> pandas.DataFrame:
        # 日付の一覧を生成
        if len(date_index1) > 0 and len(date_index2) > 0:
            start_date = min(date_index1[0], date_index2[0])
            end_date = max(date_index1[-1], date_index2[-1])
        elif len(date_index1) > 0 and len(date_index2) == 0:
            start_date = date_index1[0]
            end_date = date_index1[-1]
        elif len(date_index1) == 0 and len(date_index2) > 0:
            start_date = date_index2[0]
            end_date = date_index2[-1]
        else:
            return pandas.DataFrame()

        return pandas.DataFrame(index=_get_date_list(start_date, end_date))

    @staticmethod
    def _create_dataframe_per_date(df_task: pandas.DataFrame, df_labor: pandas.DataFrame) -> pandas.DataFrame:
        df_sub_task = df_task[
            [
                "task_id",
                "task_completed_datetime",
                "input_data_count",
                "annotation_count",
                "sum_worktime_hour",
                "annotation_worktime_hour",
                "inspection_worktime_hour",
                "acceptance_worktime_hour",
            ]
        ].copy()
        df_sub_task["task_completed_date"] = df_sub_task["task_completed_datetime"].map(
            lambda e: datetime_to_date(e) if not pandas.isna(e) else None
        )

        # タスク完了日を軸にして集計する
        value_columns = [
            "input_data_count",
            "annotation_count",
            "sum_worktime_hour",
            "annotation_worktime_hour",
            "inspection_worktime_hour",
            "acceptance_worktime_hour",
        ]
        df_agg_sub_task = df_sub_task.pivot_table(
            values=value_columns,
            index="task_completed_date",
            aggfunc=numpy.sum,
        ).fillna(0)
        if len(df_agg_sub_task) > 0:
            df_agg_sub_task["task_count"] = df_sub_task.pivot_table(
                values=["task_id"], index="task_completed_date", aggfunc="count"
            ).fillna(0)
        else:
            # 列だけ作る
            df_agg_sub_task = df_agg_sub_task.assign(**{key: 0 for key in value_columns}, task_count=0)

        if len(df_labor) > 0:
            df_labor2 = df_labor[df_labor["worktime_result_hour"] > 0]
            df_agg_labor = df_labor2.pivot_table(
                values=["worktime_result_hour"], index="date", aggfunc=numpy.sum
            ).fillna(0)
            df_tmp = df_labor2.pivot_table(values=["user_id"], index="date", aggfunc="count").fillna(0)

            if len(df_tmp) > 0:
                df_agg_labor["working_user_count"] = df_tmp
            else:
                df_agg_labor["working_user_count"] = 0

        else:
            df_agg_labor = pandas.DataFrame(columns=["worktime_result_hour", "working_user_count"])

        # 日付の一覧を生成
        df_date_base = Table._create_date_df(df_agg_sub_task.index, df_agg_labor.index)
        df_date = df_date_base.join(df_agg_sub_task).join(df_agg_labor).fillna(0)

        df_date.rename(
            columns={
                "sum_worktime_hour": "monitored_worktime_hour",
                "annotation_worktime_hour": "monitored_annotation_worktime_hour",
                "inspection_worktime_hour": "monitored_inspection_worktime_hour",
                "acceptance_worktime_hour": "monitored_acceptance_worktime_hour",
                "worktime_result_hour": "actual_worktime_hour",
                "user_id": "working_user_count",
            },
            inplace=True,
        )
        df_date["date"] = df_date.index
        return df_date

    @staticmethod
    def create_whole_productivity_per_date(df_task: pandas.DataFrame, df_labor: pandas.DataFrame) -> pandas.DataFrame:
        """
        日毎の全体の生産量、生産性を算出する。
        """
        df_date = Table._create_dataframe_per_date(df_task, df_labor)
        Table._add_ratio_cumsum_column_for_productivity_per_date(df_date)
        return df_date

    @staticmethod
    def _add_ratio_cumsum_column_for_productivity_per_date(df: pandas.DataFrame) -> pandas.DataFrame:
        """
        日毎の全体の生産量から、累計情報、生産性の列を追加する。
        """

        def add_cumsum_column(df: pandas.DataFrame, column: str):
            """累積情報の列を追加"""
            df[f"cumsum_{column}"] = df[column].cumsum()

        def add_velocity_column(df: pandas.DataFrame, numerator_column: str, denominator_column: str):
            """速度情報の列を追加"""
            MOVING_WINDOW_SIZE = 7
            MIN_WINDOW_SIZE = 2
            df[f"{numerator_column}/{denominator_column}"] = df[numerator_column] / df[denominator_column]
            df[f"{numerator_column}/{denominator_column}__lastweek"] = (
                df[numerator_column].rolling(MOVING_WINDOW_SIZE, min_periods=MIN_WINDOW_SIZE).sum()
                / df[denominator_column].rolling(MOVING_WINDOW_SIZE, min_periods=MIN_WINDOW_SIZE).sum()
            )

        # 累計情報を追加
        add_cumsum_column(df, column="task_count")
        add_cumsum_column(df, column="input_data_count")
        add_cumsum_column(df, column="actual_worktime_hour")

        # annofab 計測時間から算出したvelocityを追加
        add_velocity_column(df, numerator_column="monitored_worktime_hour", denominator_column="task_count")
        add_velocity_column(df, numerator_column="monitored_worktime_hour", denominator_column="input_data_count")
        add_velocity_column(df, numerator_column="monitored_worktime_hour", denominator_column="annotation_count")

        # 実績作業時間から算出したvelocityを追加
        add_velocity_column(df, numerator_column="actual_worktime_hour", denominator_column="task_count")
        add_velocity_column(df, numerator_column="actual_worktime_hour", denominator_column="input_data_count")
        add_velocity_column(df, numerator_column="actual_worktime_hour", denominator_column="annotation_count")

        # CSVには"INF"という文字を出力したくないので、"INF"をNaNに置換する
        df.replace([numpy.inf, -numpy.inf], numpy.nan, inplace=True)
        return df

    @staticmethod
    def merge_whole_productivity_per_date(df1: pandas.DataFrame, df2: pandas.DataFrame) -> pandas.DataFrame:
        """
        日毎の全体の生産量、生産性が格納されたDataFrameを結合する。

        Args:
            df1:
            df2:

        Returns:
            マージ済のユーザごとの生産性・品質情報
        """

        def merge_row(
            str_date: str, columns: pandas.Index, row1: Optional[pandas.Series], row2: Optional[pandas.Series]
        ) -> pandas.Series:
            if row1 is not None and row2 is not None:
                sum_row = row1.fillna(0) + row2.fillna(0)
            elif row1 is not None and row2 is None:
                sum_row = row1.fillna(0)
            elif row1 is None and row2 is not None:
                sum_row = row2.fillna(0)
            else:
                sum_row = pandas.Series(index=columns)

            sum_row.name = str_date
            return sum_row

        def date_range():
            lower_date = min(df1["date"].min(), df2["date"].min())
            upper_date = max(df1["date"].max(), df2["date"].max())
            return pandas.date_range(start=lower_date, end=upper_date)

        tmp_df1 = df1.set_index("date")
        tmp_df2 = df2.set_index("date")

        row_list: List[pandas.Series] = []
        for dt in date_range():
            str_date = str(dt.date())
            if str_date in tmp_df1.index:
                row1 = tmp_df1.loc[str_date]
            else:
                row1 = None
            if str_date in tmp_df2.index:
                row2 = tmp_df2.loc[str_date]
            else:
                row2 = None

            sum_row = merge_row(str_date=str_date, columns=tmp_df1.columns, row1=row1, row2=row2)
            row_list.append(sum_row)

        sum_df = pandas.DataFrame(row_list)
        sum_df.index.name = "date"
        Table._add_ratio_cumsum_column_for_productivity_per_date(sum_df)
        return sum_df.reset_index()



import argparse
import collections
import logging
import zipfile
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Any, Counter, Dict, Iterator, List, Optional, Set, Tuple

import pandas
from annofabapi.models import AdditionalDataDefinitionType, TaskPhase, TaskStatus
from annofabapi.parser import (
    SimpleAnnotationParser,
    SimpleAnnotationParserByTask,
    lazy_parse_simple_annotation_dir,
    lazy_parse_simple_annotation_dir_by_task,
    lazy_parse_simple_annotation_zip,
    lazy_parse_simple_annotation_zip_by_task,
)
from dataclasses_json import DataClassJsonMixin

import annofabcli
import annofabcli.common.cli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import (
    AbstractCommandLineInterface,
    ArgumentParser,
    build_annofabapi_resource_and_login,
    get_json_from_args,
    get_wait_options_from_args,
)
from annofabcli.common.dataclasses import WaitOptions
from annofabcli.common.download import DownloadingFile
from annofabcli.common.visualize import AddProps, MessageLocale

DEFAULT_WAIT_OPTIONS = WaitOptions(interval=60, max_tries=360)

logger = logging.getLogger(__name__)

AttributesColumn = Tuple[str, str, str]
"""
属性値ごとの個数を表したCSVの列の型
Tuple[Label, Attribute, Choice]
"""

LabelColumnList = List[str]

AttributesColumnList = List[AttributesColumn]


class GroupBy(Enum):
    TASK_ID = "task_id"
    INPUT_DATA_ID = "input_data_id"


@dataclass(frozen=True)
class AnnotationCounterByTask(DataClassJsonMixin):
    """"""

    task_id: str
    task_status: TaskStatus
    task_phase: TaskPhase
    task_phase_stage: int
    labels_count: Counter[str]
    attirbutes_count: Counter[AttributesColumn]


@dataclass(frozen=True)
class AnnotationCounterByInputData(DataClassJsonMixin):
    """"""

    task_id: str
    task_status: TaskStatus
    task_phase: TaskPhase
    task_phase_stage: int

    input_data_id: str
    input_data_name: str
    labels_count: Counter[str]
    attirbutes_count: Counter[AttributesColumn]


class ListAnnotationCount(AbstractCommandLineInterface):
    """
    アノテーション数情報を出力する。
    """

    CSV_FORMAT = {"encoding": "utf_8_sig", "index": False}

    @staticmethod
    def lazy_parse_simple_annotation_by_input_data(annotation_path: Path) -> Iterator[SimpleAnnotationParser]:
        if not annotation_path.exists():
            raise RuntimeError("'--annotation'で指定したディレクトリまたはファイルが存在しません。")

        if annotation_path.is_dir():
            return lazy_parse_simple_annotation_dir(annotation_path)
        elif zipfile.is_zipfile(str(annotation_path)):
            return lazy_parse_simple_annotation_zip(annotation_path)
        else:
            raise RuntimeError(f"'--annotation'で指定した'{annotation_path}'は、zipファイルまたはディレクトリではありませんでした。")

    @staticmethod
    def lazy_parse_simple_annotation_by_task(annotation_path: Path) -> Iterator[SimpleAnnotationParserByTask]:
        if not annotation_path.exists():
            raise RuntimeError("'--annotation'で指定したディレクトリまたはファイルが存在しません。")

        if annotation_path.is_dir():
            return lazy_parse_simple_annotation_dir_by_task(annotation_path)
        elif zipfile.is_zipfile(str(annotation_path)):
            return lazy_parse_simple_annotation_zip_by_task(annotation_path)
        else:
            raise RuntimeError(f"'--annotation'で指定した'{annotation_path}'は、zipファイルまたはディレクトリではありませんでした。")

    @staticmethod
    def count_for_input_data(
        simple_annotation: Dict[str, Any], target_attributes: Optional[Set[Tuple[str, str]]] = None
    ) -> AnnotationCounterByInputData:
        """
        1個の入力データに対してアノテーション数をカウントする

        Args:
            simple_annotation: JSONファイルの内容
            target_attributes:

        Returns:

        """
        details = simple_annotation["details"]
        labels_count = collections.Counter([e["label"] for e in details])

        attributes_list: List[Tuple[str, str, str]] = []
        for detail in details:
            label = detail["label"]
            for attribute, value in detail["attributes"].items():
                if target_attributes is not None and (label, attribute) in target_attributes:
                    attributes_list.append((label, attribute, str(value)))

        attirbutes_count = collections.Counter(attributes_list)

        return AnnotationCounterByInputData(
            task_id=simple_annotation["task_id"],
            task_phase=TaskPhase(simple_annotation["task_phase"]),
            task_phase_stage=simple_annotation["task_phase_stage"],
            task_status=TaskStatus(simple_annotation["task_status"]),
            input_data_id=simple_annotation["input_data_id"],
            input_data_name=simple_annotation["input_data_name"],
            labels_count=labels_count,
            attirbutes_count=attirbutes_count,
        )

    @staticmethod
    def count_for_task(
        task_parser: SimpleAnnotationParserByTask, target_attributes: Optional[Set[Tuple[str, str]]] = None
    ) -> AnnotationCounterByTask:
        """
        1個のタスクに対してアノテーション数をカウントする

        """

        labels_count: Counter[str] = collections.Counter()
        attirbutes_count: Counter[Tuple[str, str, str]] = collections.Counter()

        last_simple_annotation = None
        for parser in task_parser.lazy_parse():
            # parse()メソッドは遅いので、使わない
            simple_annotation_dict = parser.load_json()
            input_data = ListAnnotationCount.count_for_input_data(simple_annotation_dict, target_attributes)
            labels_count += input_data.labels_count
            attirbutes_count += input_data.attirbutes_count
            last_simple_annotation = simple_annotation_dict

        if last_simple_annotation is None:
            raise RuntimeError(f"{task_parser.task_id} ディレクトリにはjsonファイルが１つも含まれていません。")

        return AnnotationCounterByTask(
            task_id=last_simple_annotation["task_id"],
            task_status=TaskStatus(last_simple_annotation["task_status"]),
            task_phase=TaskPhase(last_simple_annotation["task_phase"]),
            task_phase_stage=last_simple_annotation["task_phase_stage"],
            labels_count=labels_count,
            attirbutes_count=attirbutes_count,
        )

    def print_labels_count_for_task(
        self, task_counter_list: List[AnnotationCounterByTask], label_columns: List[str], output_dir: Path
    ):
        def to_dict(c: AnnotationCounterByTask) -> Dict[str, Any]:
            d = {
                "task_id": c.task_id,
                "task_status": c.task_status.value,
                "task_phase": c.task_phase.value,
                "task_phase_stage": c.task_phase_stage,
            }
            d.update({f"label_{label}": c.labels_count[label] for label in label_columns})
            return d

        columns = ["task_id", "task_status", "task_phase", "task_phase_stage"]
        columns.extend([f"label_{e}" for e in label_columns])

        df = pandas.DataFrame([to_dict(e) for e in task_counter_list], columns=columns)
        output_file = str(output_dir / "labels_count.csv")
        annofabcli.utils.print_csv(df, output=output_file, to_csv_kwargs=self.CSV_FORMAT)

    def print_attirbutes_count_for_task(
        self,
        task_counter_list: List[AnnotationCounterByTask],
        attribute_columns: List[Tuple[str, str, str]],
        output_dir: Path,
    ):
        def to_cell(c: AnnotationCounterByTask) -> Dict[AttributesColumn, Any]:
            cell = {
                ("", "", "task_id"): c.task_id,
                ("", "", "task_status"): c.task_status.value,
                ("", "", "task_phase"): c.task_phase.value,
                ("", "", "task_phase_stage"): c.task_phase_stage,
            }
            for col in attribute_columns:
                cell.update({col: c.attirbutes_count[col]})

            return cell

        columns = [("", "", "task_id"), ("", "", "task_status"), ("", "", "task_phase"), ("", "", "task_phase_stage")]
        columns.extend(attribute_columns)
        df = pandas.DataFrame([to_cell(e) for e in task_counter_list], columns=pandas.MultiIndex.from_tuples(columns))

        output_file = str(output_dir / "attirbutes_count.csv")
        annofabcli.utils.print_csv(df, output=output_file, to_csv_kwargs=self.CSV_FORMAT)

    def print_labels_count_for_input_data(
        self, input_data_counter_list: List[AnnotationCounterByInputData], label_columns: List[str], output_dir: Path
    ):
        def to_dict(c: AnnotationCounterByInputData) -> Dict[str, Any]:
            d = {
                "input_data_id": c.input_data_id,
                "input_data_name": c.input_data_name,
                "task_id": c.task_id,
                "task_status": c.task_status.value,
                "task_phase": c.task_phase.value,
                "task_phase_stage": c.task_phase_stage,
            }
            d.update({f"label_{label}": c.labels_count[label] for label in label_columns})
            return d

        columns = ["input_data_id", "input_data_name", "task_id", "task_status", "task_phase", "task_phase_stage"]
        columns.extend([f"label_{e}" for e in label_columns])

        df = pandas.DataFrame([to_dict(e) for e in input_data_counter_list])
        output_file = str(output_dir / "labels_count.csv")
        annofabcli.utils.print_csv(df, output=output_file, to_csv_kwargs=self.CSV_FORMAT)

    def print_attirbutes_count_for_input_data(
        self,
        input_data_counter_list: List[AnnotationCounterByInputData],
        attribute_columns: List[Tuple[str, str, str]],
        output_dir: Path,
    ):
        def to_cell(c: AnnotationCounterByInputData) -> Dict[AttributesColumn, Any]:
            cell = {
                ("", "", "input_data_id"): c.input_data_id,
                ("", "", "input_data_name"): c.input_data_name,
                ("", "", "task_id"): c.task_id,
                ("", "", "task_status"): c.task_status.value,
                ("", "", "task_phase"): c.task_phase.value,
                ("", "", "task_phase_stage"): c.task_phase_stage,
            }
            for col in attribute_columns:
                cell.update({col: c.attirbutes_count[col]})

            return cell

        columns = [
            ("", "", "input_data_id"),
            ("", "", "input_data_name"),
            ("", "", "task_id"),
            ("", "", "task_status"),
            ("", "", "task_phase"),
            ("", "", "task_phase_stage"),
        ]
        columns.extend(attribute_columns)
        df = pandas.DataFrame(
            [to_cell(e) for e in input_data_counter_list], columns=pandas.MultiIndex.from_tuples(columns)
        )

        output_file = str(output_dir / "attirbutes_count.csv")
        annofabcli.utils.print_csv(df, output=output_file, to_csv_kwargs=self.CSV_FORMAT)

    @staticmethod
    def get_target_attributes_columns(annotation_specs_labels: List[Dict[str, Any]]) -> List[AttributesColumn]:
        """
        出力対象の属性情報を取得する（label, attribute, choice)
        """

        target_attributes_columns: List[AttributesColumn] = []
        for label in annotation_specs_labels:
            label_name_en = AddProps.get_message(label["label_name"], MessageLocale.EN)
            label_name_en = label_name_en if label_name_en is not None else ""

            for attribute in label["additional_data_definitions"]:
                attribute_name_en = AddProps.get_message(attribute["name"], MessageLocale.EN)
                attribute_name_en = attribute_name_en if attribute_name_en is not None else ""

                if AdditionalDataDefinitionType(attribute["type"]) in [
                    AdditionalDataDefinitionType.CHOICE,
                    AdditionalDataDefinitionType.SELECT,
                ]:
                    for choice in attribute["choices"]:
                        choice_name_en = AddProps.get_message(choice["name"], MessageLocale.EN)
                        choice_name_en = choice_name_en if choice_name_en is not None else ""
                        target_attributes_columns.append((label_name_en, attribute_name_en, choice_name_en))

                elif AdditionalDataDefinitionType(attribute["type"]) == AdditionalDataDefinitionType.FLAG:
                    target_attributes_columns.append((label_name_en, attribute_name_en, "True"))
                    target_attributes_columns.append((label_name_en, attribute_name_en, "False"))

                else:
                    continue

        return target_attributes_columns

    @staticmethod
    def get_target_label_columns(annotation_specs_labels: List[Dict[str, Any]]) -> List[str]:
        """
        出力対象の属性情報を取得する（label, attribute, choice)
        """

        def to_label_name(label: Dict[str, Any]) -> str:
            label_name_en = AddProps.get_message(label["label_name"], MessageLocale.EN)
            label_name_en = label_name_en if label_name_en is not None else ""
            return label_name_en

        return [to_label_name(label) for label in annotation_specs_labels]

    def get_target_columns(self, project_id: str) -> Tuple[LabelColumnList, AttributesColumnList]:
        annotation_specs, _ = self.service.api.get_annotation_specs(project_id)
        annotation_specs_labels = annotation_specs["labels"]
        label_columns = self.get_target_label_columns(annotation_specs_labels)
        attributes_columns = self.get_target_attributes_columns(annotation_specs_labels)
        return (label_columns, attributes_columns)

    def list_annotation_count_by_task(self, project_id: str, annotation_path: Path, output_dir: Path) -> None:
        task_counter_list = []
        iter_task_parser = self.lazy_parse_simple_annotation_by_task(annotation_path)
        target_label_columns, target_attributes_columns = self.get_target_columns(project_id)

        target_attributes = {(e[0], e[1]) for e in target_attributes_columns}
        logger.info(f"アノテーションzip/ディレクトリを読み込み中")
        for task_index, task_parser in enumerate(iter_task_parser):
            task_index += 1
            if task_index % 1000 == 0:
                logger.debug(f"{task_index}  件目を読み込み中")

            task_counter = self.count_for_task(task_parser, target_attributes=target_attributes)
            task_counter_list.append(task_counter)

        self.print_labels_count_for_task(task_counter_list, label_columns=target_label_columns, output_dir=output_dir)

        self.print_attirbutes_count_for_task(
            task_counter_list,
            output_dir=output_dir,
            attribute_columns=target_attributes_columns,
        )

    def list_annotation_count_by_input_data(self, project_id: str, annotation_path: Path, output_dir: Path) -> None:
        input_data_counter_list = []
        iter_parser = self.lazy_parse_simple_annotation_by_input_data(annotation_path)
        target_label_columns, target_attributes_columns = self.get_target_columns(project_id)

        target_attributes = {(e[0], e[1]) for e in target_attributes_columns}
        logger.info(f"アノテーションzip/ディレクトリを読み込み中")
        for index, parser in enumerate(iter_parser):
            if index % 1000 == 0:
                logger.debug(f"{index}  件目を読み込み中")

            simple_annotation_dict = parser.load_json()
            input_data_counter = self.count_for_input_data(simple_annotation_dict, target_attributes=target_attributes)
            input_data_counter_list.append(input_data_counter)

        self.print_labels_count_for_input_data(
            input_data_counter_list, label_columns=target_label_columns, output_dir=output_dir
        )

        self.print_attirbutes_count_for_input_data(
            input_data_counter_list,
            output_dir=output_dir,
            attribute_columns=target_attributes_columns,
        )

    def main(self):
        args = self.args

        project_id = args.project_id
        super().validate_project(project_id, project_member_roles=None)

        if args.annotation is not None:
            annotation_path = Path(args.annotation)
        else:
            cache_dir = annofabcli.utils.get_cache_dir()
            annotation_path = cache_dir / f"annotation-{project_id}.zip"
            wait_options = get_wait_options_from_args(get_json_from_args(args.wait_options), DEFAULT_WAIT_OPTIONS)
            downloading_obj = DownloadingFile(self.service)
            downloading_obj.download_annotation_zip(
                project_id,
                dest_path=str(annotation_path),
                is_latest=args.latest,
                wait_options=wait_options,
            )

        group_by = GroupBy(args.group_by)
        if group_by == GroupBy.TASK_ID:
            self.list_annotation_count_by_task(
                project_id, annotation_path=annotation_path, output_dir=Path(args.output_dir)
            )
        elif group_by == GroupBy.INPUT_DATA_ID:
            self.list_annotation_count_by_input_data(
                project_id, annotation_path=annotation_path, output_dir=Path(args.output_dir)
            )


def parse_args(parser: argparse.ArgumentParser):
    argument_parser = ArgumentParser(parser)

    argument_parser.add_project_id()
    parser.add_argument(
        "--annotation", type=str, help="アノテーションzip、またはzipを展開したディレクトリを指定します。" "指定しない場合はAnnoFabからダウンロードします。"
    )
    parser.add_argument("-o", "--output_dir", type=str, required=True, help="出力ディレクトリのパス")

    parser.add_argument(
        "--group_by",
        type=str,
        choices=[GroupBy.TASK_ID.value, GroupBy.INPUT_DATA_ID.value],
        default=GroupBy.TASK_ID.value,
        help="アノテーションの個数をどの単位で集約するかを指定してます。デフォルトは'task_id'です。",
    )

    parser.add_argument(
        "--latest",
        action="store_true",
        help="'--annotation'を指定しないとき、最新のアノテーションzipを参照します。このオプションを指定すると、アノテーションzipを更新するのに数分待ちます。",
    )

    parser.add_argument(
        "--wait_options",
        type=str,
        help="アノテーションzipの更新が完了するまで待つ際のオプションを、JSON形式で指定してください。"
        "`file://`を先頭に付けるとjsonファイルを指定できます。"
        'デフォルは`{"interval":60, "max_tries":360}` です。'
        "`interval`:完了したかを問い合わせる間隔[秒], "
        "`max_tires`:完了したかの問い合わせを最大何回行うか。",
    )

    parser.set_defaults(subcommand_func=main)


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    ListAnnotationCount(service, facade, args).main()


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "list_annotation_count"
    subcommand_help = "各ラベル、各属性値のアノテーション数を出力します。"
    description = "各ラベル、各属性値のアノテーション数を、タスクごと/入力データごとに出力します。"
    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description=description)
    parse_args(parser)



import argparse
import datetime
import logging
import sys
from pathlib import Path
from typing import Any, Dict, List, Optional  # pylint: disable=unused-import

import annofabapi
import numpy as np
import pandas as pd
from annofabapi.models import ProjectMemberRole

import annofabcli
import annofabcli.common.cli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import (
    AbstractCommandLineInterface,
    ArgumentParser,
    build_annofabapi_resource_and_login,
    get_list_from_args,
)
from annofabcli.experimental.utils import (
    FormatTarget,
    TimeUnitTarget,
    add_id_csv,
    create_column_list,
    create_column_list_per_project,
    print_byname_total_list,
    print_time_list_from_work_time_list,
    print_total,
    timeunit_conversion,
)

logger = logging.getLogger(__name__)


class Database:
    def __init__(
        self,
        annofab_service: annofabapi.Resource,
        project_id: str,
        organization_id: str,
        start_date: str,
        end_date: str,
    ):
        self.annofab_service = annofab_service
        self.project_id = project_id
        self.organization_id = organization_id
        self.start_date = start_date
        self.end_date = end_date

    def get_labor_control(self) -> List[Dict[str, Any]]:
        labor_control_df = self.annofab_service.api.get_labor_control(
            {
                "organization_id": self.organization_id,
                "project_id": self.project_id,
                "from": self.start_date,
                "to": self.end_date,
            }
        )[0]
        return labor_control_df

    def get_account_statistics(self) -> List[Dict[str, Any]]:
        account_statistics = self.annofab_service.wrapper.get_account_statistics(self.project_id)
        return account_statistics

    def get_project_members(self) -> dict:
        project_members = self.annofab_service.api.get_project_members(project_id=self.project_id)[0]
        return project_members["list"]


class Table:
    def __init__(self, database: Database, facade: AnnofabApiFacade):
        self.database = database
        self.project_id = database.project_id
        self.facade = facade

    def create_labor_control_df(self):

        labor_control = self.database.get_labor_control()
        labor_control_list = []
        for labor in labor_control:
            if labor["account_id"] is not None:
                new_history = {
                    "user_name": self._get_username(labor["account_id"]),
                    "user_id": self._get_user_id(labor["account_id"]),
                    "user_biography": self._get_user_biography(labor["account_id"]),
                    "date": labor["date"],
                    "worktime_planned": np.nan
                    if labor["values"]["working_time_by_user"]["plans"] is None
                    else int(labor["values"]["working_time_by_user"]["plans"]) / 60000,
                    "worktime_actual": np.nan
                    if labor["values"]["working_time_by_user"]["results"] is None
                    else int(labor["values"]["working_time_by_user"]["results"]) / 60000,
                    "working_description": labor["values"]["working_time_by_user"]["description"]
                    if labor["values"]["working_time_by_user"]["results"] is not None
                    else None,
                }
                labor_control_list.append(new_history)

        return labor_control_list

    def create_account_statistics_df(self):
        """
        メンバごと、日ごとの作業時間
        """
        account_statistics = self.database.get_account_statistics()
        all_histories = []
        for account_info in account_statistics:
            account_id = account_info["account_id"]
            histories = account_info["histories"]
            if account_id is not None:
                for history in histories:
                    new_history = {
                        "user_name": self._get_username(account_id),
                        "user_id": self._get_user_id(account_id),
                        "user_biography": self._get_user_biography(account_id),
                        "date": history["date"],
                        "worktime_monitored": annofabcli.utils.isoduration_to_minute(history["worktime"]),
                    }
                    all_histories.append(new_history)

        return all_histories

    def create_afaw_time_df(self) -> pd.DataFrame:
        account_statistics_df = pd.DataFrame(self.create_account_statistics_df())
        labor_control_df = pd.DataFrame(self.create_labor_control_df())
        if len(account_statistics_df) == 0 and len(labor_control_df) == 0:
            df = pd.DataFrame([])
        elif len(account_statistics_df) == 0:

            labor_control_df["worktime_monitored"] = np.nan
            df = labor_control_df
        elif len(labor_control_df) == 0:
            account_statistics_df["worktime_planned"] = np.nan
            account_statistics_df["worktime_actual"] = np.nan
            df = account_statistics_df
        else:
            df = pd.merge(
                account_statistics_df,
                labor_control_df,
                on=["user_name", "user_id", "date", "user_biography"],
                how="outer",
            )
        df["project_id"] = self.project_id
        df["project_title"] = self.facade.get_project_title(self.project_id)
        return df

    def _get_user_id(self, account_id: Optional[str]) -> Optional[str]:
        """
        プロジェクトメンバのuser_idを取得する。プロジェクトメンバでなければ、account_idを返す。
        account_idがNoneならばNoneを返す。
        """
        if account_id is None:
            return None

        member = self.facade.get_project_member_from_account_id(self.project_id, account_id)
        if member is not None:
            return member["user_id"]
        else:
            return account_id

    def _get_username(self, account_id: Optional[str]) -> Optional[str]:
        """
        プロジェクトメンバのusernameを取得する。プロジェクトメンバでなければ、account_idを返す。
        account_idがNoneならばNoneを返す。
        """
        if account_id is None:
            return None

        member = self.facade.get_project_member_from_account_id(self.project_id, account_id)
        if member is not None:
            return member["username"]
        else:
            return account_id

    def _get_user_biography(self, account_id: Optional[str]) -> Optional[str]:
        """
        プロジェクトメンバのbiographyを取得する。プロジェクトメンバでなければ、account_idを返す。
        account_idがNoneならばNoneを返す。
        """
        if account_id is None:
            return None

        member = self.facade.get_project_member_from_account_id(self.project_id, account_id)
        if member is not None:
            return member["biography"]
        else:
            return account_id


def get_organization_id_from_project_id(annofab_service: annofabapi.Resource, project_id: str) -> str:
    """
    project_idからorganization_idを返す
    """
    organization, _ = annofab_service.api.get_organization_of_project(project_id)
    return organization["organization_id"]


def refine_df(
    df: pd.DataFrame, start_date: datetime.date, end_date: datetime.date, user_id_list: Optional[List[str]]
) -> pd.DataFrame:
    # 日付で絞り込み
    df["date"] = pd.to_datetime(df["date"]).dt.date
    refine_day_df = df[(df["date"] >= start_date) & (df["date"] <= end_date)].copy()
    # user_id を絞り込み
    refine_user_df = (
        refine_day_df
        if user_id_list is None
        else refine_day_df[refine_day_df["user_id"].str.contains("|".join(user_id_list), case=False)].copy()
    )
    return refine_user_df


class ListLaborWorktime(AbstractCommandLineInterface):
    """
    労務管理画面の作業時間を出力する
    """

    def _get_project_title_list(self, project_id_list: List[str]) -> List[str]:
        return [self.facade.get_project_title(project_id) for project_id in project_id_list]

    def list_labor_worktime(self, project_id: str, start_date: str, end_date: str) -> pd.DataFrame:

        """"""

        super().validate_project(
            project_id, project_member_roles=[ProjectMemberRole.OWNER, ProjectMemberRole.TRAINING_DATA_USER]
        )
        # プロジェクト or 組織に対して、必要な権限が付与されているかを確認

        organization_id = get_organization_id_from_project_id(self.service, project_id)
        database = Database(self.service, project_id, organization_id, start_date, end_date=end_date)
        # Annofabから取得した情報に関するデータベースを取得するクラス
        table_obj = Table(database=database, facade=self.facade)
        # Databaseから取得した情報を元にPandas DataFrameを生成するクラス
        #     チェックポイントファイルがあること前提
        return table_obj.create_afaw_time_df()

    def _output(self, output: Any, df: pd.DataFrame, index: bool, add_project_id: bool, project_id_list: List[str]):
        if isinstance(output, str):
            Path(output).parent.mkdir(exist_ok=True, parents=True)
        df.to_csv(
            output,
            date_format="%Y-%m-%d",
            encoding="utf_8_sig",
            line_terminator="\r\n",
            float_format="%.2f",
            index=index,
        )
        if output != sys.stdout and add_project_id:
            add_id_csv(output, self._get_project_title_list(project_id_list))

    def main(self):
        args = self.args
        format_target = FormatTarget(args.format)
        time_unit = TimeUnitTarget(args.time_unit)

        start_date = datetime.datetime.strptime(args.start_date, "%Y-%m-%d").date()
        end_date = datetime.datetime.strptime(args.end_date, "%Y-%m-%d").date()
        user_id_list = get_list_from_args(args.user_id) if args.user_id is not None else None

        total_df = pd.DataFrame([])

        # プロジェクトごとにデータを取得
        project_id_list = get_list_from_args(args.project_id)
        logger.info(f"{len(project_id_list)} 件のプロジェクトを取得します。")
        for i, project_id in enumerate(list(set(project_id_list))):
            logger.debug(f"{i + 1} 件目: project_id = {project_id}")
            try:
                afaw_time_df = self.list_labor_worktime(
                    project_id, start_date.strftime("%Y-%m-%d"), end_date.strftime("%Y-%m-%d")
                )
                total_df = pd.concat([total_df, afaw_time_df], sort=True)
            except Exception as e:  # pylint: disable=broad-except
                logger.error(e)
                logger.error(f"プロジェクトにアクセスできませんでした（project_id={project_id} ）。")

        # データが無い場合にはwarning
        if len(total_df) == 0:
            logger.warning(f"対象プロジェクトの労務管理情報・作業情報が0件のため、出力しません。")
            return
        total_df = refine_df(total_df, start_date, end_date, user_id_list)
        if len(total_df) == 0:
            logger.warning(f"対象期間の労務管理情報・作業情報が0件のため、出力しません。")
            return

        # 時間単位変換
        total_df = timeunit_conversion(df=total_df, time_unit=time_unit)

        # フォーマット別に出力dfを作成
        if format_target == FormatTarget.BY_NAME_TOTAL:
            df = print_byname_total_list(total_df)
        elif format_target == FormatTarget.TOTAL:
            df = print_total(total_df)
        elif format_target == FormatTarget.COLUMN_LIST_PER_PROJECT:
            df = create_column_list_per_project(total_df)
        elif format_target == FormatTarget.COLUMN_LIST:
            df = create_column_list(total_df)
        elif format_target == FormatTarget.DETAILS:
            df = print_time_list_from_work_time_list(total_df)
        else:
            raise RuntimeError(f"format_target='{format_target}'は対象外です。")
        # 出力先別に出力
        if args.output:
            out_format = args.output
        else:
            out_format = sys.stdout
        self._output(
            out_format,
            df,
            index=(format_target == FormatTarget.DETAILS),
            add_project_id=args.add_project_id,
            project_id_list=project_id_list,
        )


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    ListLaborWorktime(service, facade, args).main()


def parse_args(parser: argparse.ArgumentParser):
    argument_parser = ArgumentParser(parser)
    time_unit_choices = [e.value for e in TimeUnitTarget]
    format_choices = [e.value for e in FormatTarget]
    parser.add_argument(
        "-p",
        "--project_id",
        type=str,
        required=True,
        nargs="+",
        help="集計対象のプロジェクトのproject_idを指定します。複数指定した場合は合計値を出力します。`file://`を先頭に付けると、project_idの一覧が記載されたファイルを指定できます。",
    )
    parser.add_argument(
        "-u",
        "--user_id",
        type=str,
        nargs="+",
        default=None,
        help="集計対象のユーザのuser_idに部分一致するものを集計します。"
        "指定しない場合は、プロジェクトメンバが指定されます。`file://`を先頭に付けると、user_idの一覧が記載されたファイルを指定できます。",
    )
    parser.add_argument("--start_date", type=str, required=True, help="集計開始日(YYYY-mm-dd)")
    parser.add_argument("--end_date", type=str, required=True, help="集計終了日(YYYY-mm-dd)")

    parser.add_argument("--time_unit", type=str, default="h", choices=time_unit_choices, help="出力の時間単位(h/m/s)")
    parser.add_argument(
        "-f",
        "--format",
        type=str,
        choices=format_choices,
        default="details",
        help="出力する際のフォーマットです。デフォルトは'details'です。"
        "details:日毎・人毎の詳細な値を出力する, "
        "total:期間中の合計値だけを出力する, "
        "by_name_total:人毎の集計の合計値を出力する, "
        "column_list:列固定で詳細な値を出力する, "
        "column_list_per_project:列固定で詳細な値をプロジェクトごとに出力する, ",
    )
    parser.add_argument("--add_project_id", action="store_true", help="出力する際にprojectidを出力する")
    argument_parser.add_output(required=False)

    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "list_labor_worktime"
    subcommand_help = "労務管理画面の作業時間を出力します。"
    description = "作業者ごとに、「作業者が入力した実績時間」と「AnnoFabが集計した作業時間」の差分を出力します。"

    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description)
    parse_args(parser)



import argparse
import logging
from pathlib import Path
from typing import List, Optional

import pandas

import annofabcli
from annofabcli.common.cli import get_list_from_args
from annofabcli.common.utils import print_csv, print_json
from annofabcli.experimental.merge_peformance_per_date import create_df_merged_peformance_per_date
from annofabcli.experimental.merge_peformance_per_user import create_df_merged_peformance_per_user
from annofabcli.experimental.write_linegraph_per_user import write_linegraph_per_user
from annofabcli.experimental.write_performance_scatter_per_user import write_performance_scatter_per_user
from annofabcli.experimental.write_task_histogram import write_task_histogram
from annofabcli.experimental.write_whole_linegraph import write_whole_linegraph
from annofabcli.statistics.csv import (
    FILENAME_PEFORMANCE_PER_DATE,
    FILENAME_PEFORMANCE_PER_USER,
    FILENAME_TASK_LIST,
    Csv,
)

logger = logging.getLogger(__name__)


def merge_visualization_dir(
    project_dir_list: List[Path],
    output_dir: Path,
    user_id_list: Optional[List[str]] = None,
    minimal_output: bool = False,
):
    def execute_merge_peformance_per_user():
        performance_per_user_csv_list = [dir / FILENAME_PEFORMANCE_PER_USER for dir in project_dir_list]
        df_per_user = create_df_merged_peformance_per_user(csv_path_list=performance_per_user_csv_list)
        csv_obj.write_productivity_per_user(df_per_user)
        csv_obj.write_whole_productivity(df_per_user)

    def execute_merge_peformance_per_date():
        performance_per_date_csv_list = [dir / FILENAME_PEFORMANCE_PER_DATE for dir in project_dir_list]
        df_per_date = create_df_merged_peformance_per_date(performance_per_date_csv_list)
        csv_obj.write_whole_productivity_per_date(df_per_date)

    def merge_task_list() -> pandas.DataFrame:
        list_df = []
        for project_dir in project_dir_list:
            csv_path = project_dir / FILENAME_TASK_LIST
            if csv_path.exists():
                list_df.append(pandas.read_csv(str(csv_path)))
            else:
                logger.warning(f"{csv_path} は存在しませんでした。")
                continue

        df = pandas.concat(list_df, axis=0)
        return df

    def write_csv_for_summary(df_task: pandas.DataFrame) -> None:
        """
        タスク関係のCSVを出力する。
        """
        csv_obj.write_task_count_summary(df_task)
        csv_obj.write_worktime_summary(df_task)
        csv_obj.write_count_summary(df_task)

    def write_info_json() -> None:
        info = {"target_dir_list": [str(e) for e in project_dir_list]}
        print_json(info, is_pretty=True, output=str(output_dir / "info.json"))

    # CSV生成
    csv_obj = Csv(str(output_dir))

    execute_merge_peformance_per_user()
    execute_merge_peformance_per_date()
    df_task = merge_task_list()
    print_csv(df_task, output=str(output_dir / FILENAME_TASK_LIST))
    write_csv_for_summary(df_task)

    # HTML生成
    write_performance_scatter_per_user(csv=output_dir / FILENAME_PEFORMANCE_PER_USER, output_dir=output_dir / "scatter")
    write_whole_linegraph(csv=output_dir / FILENAME_PEFORMANCE_PER_DATE, output_dir=output_dir / "line-graph")
    write_linegraph_per_user(
        csv=output_dir / FILENAME_TASK_LIST,
        output_dir=output_dir / "line-graph",
        minimal_output=minimal_output,
        user_id_list=user_id_list,
    )
    write_task_histogram(
        csv=output_dir / FILENAME_TASK_LIST, output_dir=output_dir / "histogram", minimal_output=minimal_output
    )

    # info.jsonを出力
    write_info_json()


def main(args):
    user_id_list = get_list_from_args(args.user_id) if args.user_id is not None else None
    merge_visualization_dir(
        project_dir_list=args.dir,
        user_id_list=user_id_list,
        minimal_output=args.minimal,
        output_dir=args.output_dir,
    )


def parse_args(parser: argparse.ArgumentParser):
    parser.add_argument("--dir", type=Path, nargs="+", required=True, help="マージ対象ディレクトリ")
    parser.add_argument("-o", "--output_dir", type=Path, required=True, help="出力先ディレクトリ。配下にプロジェクト名のディレクトリが出力される。")

    parser.add_argument(
        "-u",
        "--user_id",
        nargs="+",
        help=(
            "ユーザごとの折れ線グラフに表示するユーザのuser_idを指定してください。"
            "指定しない場合は、上位20人のユーザ情報がプロットされます。"
            "file://`を先頭に付けると、一覧が記載されたファイルを指定できます。"
        ),
    )

    parser.add_argument(
        "--minimal",
        action="store_true",
        help="必要最小限のファイルを出力します。",
    )

    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "merge_visualization_dir"
    subcommand_help = "`annofabcli statistics visualize`コマンドの出力結果をマージします。"
    description = "`annofabcli statistics visualize`コマンドの出力結果をマージします。"
    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description)
    parse_args(parser)



import argparse
import logging
import multiprocessing
import sys
import uuid
from functools import partial
from typing import Any, Dict, List, Optional, Tuple

import annofabapi.utils
import dateutil
import requests
from annofabapi.dataclass.task import Task
from annofabapi.models import Inspection, InspectionStatus, ProjectMemberRole, TaskPhase, TaskStatus
from more_itertools import first_true

import annofabcli
import annofabcli.common.cli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import (
    AbstracCommandCinfirmInterface,
    AbstractCommandLineInterface,
    ArgumentParser,
    build_annofabapi_resource_and_login,
)

logger = logging.getLogger(__name__)

InspectionJson = Dict[str, Dict[str, List[Inspection]]]
"""
Dict[task_id, Dict[input_data_id, List[Inspection]]] の検査コメント情報
"""


class CompleteTasksMain(AbstracCommandCinfirmInterface):
    def __init__(self, service: annofabapi.Resource, all_yes: bool = False):
        self.service = service
        self.facade = AnnofabApiFacade(service)
        AbstracCommandCinfirmInterface.__init__(self, all_yes)

    @staticmethod
    def inspection_list_to_input_data_dict(inspection_list: List[Inspection]) -> Dict[str, List[Inspection]]:
        """
        検査コメントのListを、Dict[InputDataId, List[Inspection]]] の形式に変換する。

        """
        input_data_dict: Dict[str, List[Inspection]] = {}
        for inspection in inspection_list:
            input_data_id = inspection["input_data_id"]
            inspection_list = input_data_dict.get(input_data_id, [])
            inspection_list.append(inspection)
            input_data_dict[input_data_id] = inspection_list
        return input_data_dict

    @staticmethod
    def inspection_list_to_dict(all_inspection_list: List[Inspection]) -> InspectionJson:
        """
        検査コメントのListを、Dict[TaskId, Dict[InputDataId, List[Inspection]]] の形式に変換する。

        """
        task_dict: InspectionJson = {}
        for inspection in all_inspection_list:
            task_id = inspection["task_id"]
            input_data_dict: Dict[str, List[Inspection]] = task_dict.get(task_id, {})

            input_data_id = inspection["input_data_id"]
            inspection_list = input_data_dict.get(input_data_id, [])
            inspection_list.append(inspection)
            input_data_dict[input_data_id] = inspection_list

            task_dict[task_id] = input_data_dict

        return task_dict

    def reply_inspection_comment(
        self,
        task: Task,
        input_data_id: str,
        unanswered_comment_list: List[Inspection],
        reply_comment: str,
    ):
        """
        未回答の検査コメントに対して、返信を付与する。
        """

        def to_req_inspection(i: Inspection) -> Dict[str, Any]:
            return {
                "data": {
                    "project_id": task.project_id,
                    "comment": reply_comment,
                    "task_id": task.task_id,
                    "input_data_id": input_data_id,
                    "inspection_id": str(uuid.uuid4()),
                    "phase": task.phase.value,
                    "phase_stage": task.phase_stage,
                    "commenter_account_id": self.service.api.account_id,
                    "data": i["data"],
                    "parent_inspection_id": i["inspection_id"],
                    "status": InspectionStatus.NO_CORRECTION_REQUIRED.value,
                    "created_datetime": task.updated_datetime,
                },
                "_type": "Put",
            }

        request_body = [to_req_inspection(e) for e in unanswered_comment_list]
        return self.service.api.batch_update_inspections(
            task.project_id, task.task_id, input_data_id, request_body=request_body
        )[0]

    def update_status_of_inspections(
        self,
        task: Task,
        input_data_id: str,
        inspection_list: List[Inspection],
        inspection_status: InspectionStatus,
    ):

        if inspection_list is None or len(inspection_list) == 0:
            logger.warning(f"変更対象の検査コメントはなかった。task_id = {task.task_id}, input_data_id = {input_data_id}")
            return

        target_inspection_id_list = [inspection["inspection_id"] for inspection in inspection_list]

        def filter_inspection(arg_inspection: Inspection) -> bool:
            """
            statusを変更する検査コメントの条件。
            """

            return arg_inspection["inspection_id"] in target_inspection_id_list

        self.service.wrapper.update_status_of_inspections(
            task.project_id,
            task.task_id,
            input_data_id,
            filter_inspection=filter_inspection,
            inspection_status=inspection_status,
            updated_datetime=task.updated_datetime,
        )
        logger.debug(f"{task.task_id}, {input_data_id}, {len(inspection_list)}件 検査コメントの状態を変更")

    def get_unprocessed_inspection_list(self, task: Task, input_data_id) -> List[Inspection]:
        """
        未処置の検査コメントリストを取得する。
        ただし、現在のタスクフェーズで編集できる検査コメントのみである。

        Args:
            task:
            input_data_id:
            target_phase:
            target_phase_stage:

        Returns:

        """
        inspectin_list, _ = self.service.api.get_inspections(task.project_id, task.task_id, input_data_id)
        return [
            e
            for e in inspectin_list
            if e["status"] == InspectionStatus.ANNOTATOR_ACTION_REQUIRED.value
            and e["phase"] == task.phase.value
            and e["phase_stage"] == task.phase_stage
        ]

    def get_unprocessed_inspection_list_by_task_id(
        self, project_id: str, task: Task, target_phase: TaskPhase, target_phase_stage: int
    ) -> List[Inspection]:
        all_inspection_list = []
        for input_data_id in task.input_data_id_list:
            inspectins, _ = self.service.api.get_inspections(project_id, task.task_id, input_data_id)
            all_inspection_list.extend(
                [
                    e
                    for e in inspectins
                    if e["status"] == InspectionStatus.ANNOTATOR_ACTION_REQUIRED.value
                    and e["phase"] == target_phase.value
                    and e["phase_stage"] == target_phase_stage
                ]
            )

        return all_inspection_list

    def change_to_working_status(self, task: Task) -> Task:
        """
        必要なら担当者を変更して、作業中状態にします。

        Args:
            task:

        Returns:
            作業中状態後のタスク
        """
        # 担当者変更
        my_account_id = self.service.api.account_id
        try:
            if task.account_id != my_account_id:
                self.facade.change_operator_of_task(task.project_id, task.task_id, my_account_id)
                logger.debug(f"{task.task_id}: 担当者を自分自身に変更しました。")

            dict_task = self.facade.change_to_working_status(
                project_id=task.project_id, task_id=task.task_id, account_id=my_account_id
            )
            return Task.from_dict(dict_task)

        except requests.HTTPError as e:
            logger.warning(f"{task.task_id}: 担当者の変更、または作業中状態への変更に失敗しました。")
            raise e

    def get_unanswered_comment_list(self, task: Task, input_data_id: str) -> List[Inspection]:
        """
        未回答の検査コメントのリストを取得する。

        Returns:

        """

        def exists_answered_comment(parent_inspection_id: str) -> bool:
            """
            回答済の返信コメントがあるかどうか

            Args:
                parent_inspection_id:

            Returns:

            """
            if task.started_datetime is None:
                # タスク未着手状態なので、回答済のコメントはない
                return False
            task_started_datetime = task.started_datetime
            answered_comment = first_true(
                inspection_list,
                pred=lambda e: e["parent_inspection_id"] == parent_inspection_id
                and dateutil.parser.parse(e["created_datetime"]) >= dateutil.parser.parse(task_started_datetime),
            )
            return answered_comment is not None

        inspection_list, _ = self.service.api.get_inspections(task.project_id, task.task_id, input_data_id)
        # 未処置の検査コメント
        unprocessed_inspection_list = [
            e
            for e in inspection_list
            if e["parent_inspection_id"] is None and e["status"] == InspectionStatus.ANNOTATOR_ACTION_REQUIRED.value
        ]

        unanswered_comment_list = [
            e for e in unprocessed_inspection_list if not exists_answered_comment(e["inspection_id"])
        ]
        return unanswered_comment_list

    def complete_task_for_annotation_phase(
        self,
        task: Task,
        reply_comment: Optional[str] = None,
    ) -> bool:
        """
        annotation phaseのタスクを完了状態にする。

        Args:
            project_id:
            task: 操作対象のタスク。annotation phase状態であること前提。
            reply_comment: 未処置の検査コメントに対する返信コメント。Noneの場合、スキップする。

        Returns:
            成功したかどうか
        """

        unanswered_comment_list_dict: Dict[str, List[Inspection]] = {}
        for input_data_id in task.input_data_id_list:
            unanswered_comment_list = self.get_unanswered_comment_list(task, input_data_id)
            unanswered_comment_list_dict[input_data_id] = unanswered_comment_list

        unanswered_comment_count_for_task = sum([len(e) for e in unanswered_comment_list_dict.values()])
        if unanswered_comment_count_for_task == 0:
            if not self.confirm_processing(f"タスク'{task.task_id}'の教師付フェーズを次のフェーズに進めますか？"):
                return False

            self.change_to_working_status(task)
            self.facade.complete_task(task.project_id, task.task_id)
            logger.info(f"{task.task_id}: 教師付フェーズを次のフェーズに進めました。")
            return True
        else:
            logger.debug(f"{task.task_id}: 未回答の検査コメントが {unanswered_comment_count_for_task} 件あります。")
            if reply_comment is None:
                logger.warning(f"{task.task_id}: 未回答の検査コメントに対する返信コメント（'--reply_comment'）が指定されていないので、スキップします。")
                return False
            elif not self.confirm_processing(f"タスク'{task.task_id}'の教師付フェーズを次のフェーズに進めますか？"):
                return False
            else:
                changed_task = self.change_to_working_status(task)

                logger.debug(f"{task.task_id}: 未回答の検査コメント {unanswered_comment_count_for_task} 件に対して、返信コメントを付与します。")
                for input_data_id, unanswered_comment_list in unanswered_comment_list_dict.items():
                    if len(unanswered_comment_list) == 0:
                        continue
                    self.reply_inspection_comment(
                        changed_task,
                        input_data_id=input_data_id,
                        unanswered_comment_list=unanswered_comment_list,
                        reply_comment=reply_comment,
                    )

                self.facade.complete_task(task.project_id, task.task_id)
                logger.info(f"{task.task_id}: 教師付フェーズをフェーズに進めました。")
                return True

    def complete_task_for_inspection_acceptance_phase(
        self,
        task: Task,
        inspection_status: Optional[InspectionStatus] = None,
    ) -> bool:
        unprocessed_inspection_list_dict: Dict[str, List[Inspection]] = {}
        for input_data_id in task.input_data_id_list:
            unprocessed_inspection_list = self.get_unprocessed_inspection_list(task, input_data_id)
            unprocessed_inspection_list_dict[input_data_id] = unprocessed_inspection_list

        unprocessed_inspection_count = sum([len(e) for e in unprocessed_inspection_list_dict.values()])

        if unprocessed_inspection_count == 0:
            if not self.confirm_processing(f"タスク'{task.task_id}'の検査/受入フェーズを次のフェーズに進めますか？"):
                return False

            self.change_to_working_status(task)
            self.facade.complete_task(task.project_id, task.task_id)
            logger.info(f"{task.task_id}: 検査/受入フェーズを次のフェーズに進めました。")
            return True

        else:
            logger.debug(f"{task.task_id}: 未処置の検査コメントが {unprocessed_inspection_count} 件あります。")
            if inspection_status is None:
                logger.warning(f"{task.task_id}: 未処置の検査コメントに対する対応方法（'--inspection_status'）が指定されていないので、スキップします。")
                return False
            elif not self.confirm_processing(f"タスク'{task.task_id}'の検査/受入フェーズを次のフェーズに進めますか？"):
                return False

            changed_task = self.change_to_working_status(task)

            logger.debug(f"{task.task_id}: 未処置の検査コメントを、{inspection_status.value} 状態にします。")
            for input_data_id, unprocessed_inspection_list in unprocessed_inspection_list_dict.items():
                if len(unprocessed_inspection_list) == 0:
                    continue

                self.update_status_of_inspections(
                    changed_task,
                    input_data_id,
                    inspection_list=unprocessed_inspection_list,
                    inspection_status=inspection_status,
                )

            self.facade.complete_task(task.project_id, task.task_id)
            logger.info(f"{task.task_id}: 検査/受入フェーズを次のフェーズに進めました。")
            return True

    def complete_task(
        self,
        project_id: str,
        task_id: str,
        target_phase: TaskPhase,
        target_phase_stage: int,
        reply_comment: Optional[str] = None,
        inspection_status: Optional[InspectionStatus] = None,
        task_index: Optional[int] = None,
    ) -> bool:
        logging_prefix = f"{task_index + 1} 件目" if task_index is not None else ""

        dict_task = self.service.wrapper.get_task_or_none(project_id, task_id)
        if dict_task is None:
            logger.warning(f"{task_id} のタスクを取得できませんでした。")
            return False

        task: Task = Task.from_dict(dict_task)
        logger.info(
            f"{logging_prefix} : タスク情報 task_id={task_id}, "
            f"phase={task.phase.value}, phase_stage={task.phase_stage}, status={task.status.value}"
        )
        if not (task.phase == target_phase and task.phase_stage == target_phase_stage):
            logger.warning(f"{task_id} は操作対象のフェーズ、フェーズステージではないため、スキップします。")
            return False

        if task.status == TaskStatus.COMPLETE:
            logger.warning(f"{task_id} は既に完了状態であるため、スキップします。")
            return False

        try:
            if task.phase == TaskPhase.ANNOTATION:
                return self.complete_task_for_annotation_phase(task, reply_comment=reply_comment)
            else:
                return self.complete_task_for_inspection_acceptance_phase(task, inspection_status=inspection_status)

        except Exception as e:  # pylint: disable=broad-except
            logger.warning(f"{task_id}: {task.phase} フェーズを完了状態にするのに失敗しました。")
            logger.warning(e)
            new_task: Task = Task.from_dict(self.service.wrapper.get_task_or_none(project_id, task_id))
            if new_task.status == TaskStatus.WORKING and new_task.account_id == self.service.api.account_id:
                self.facade.change_to_break_phase(project_id, task_id)
            return False

    def complete_task_for_task_wrapper(
        self,
        tpl: Tuple[int, str],
        project_id: str,
        target_phase: TaskPhase,
        target_phase_stage: int,
        reply_comment: Optional[str] = None,
        inspection_status: Optional[InspectionStatus] = None,
    ) -> bool:
        task_index, task_id = tpl
        return self.complete_task(
            project_id=project_id,
            task_id=task_id,
            task_index=task_index,
            target_phase=target_phase,
            target_phase_stage=target_phase_stage,
            reply_comment=reply_comment,
            inspection_status=inspection_status,
        )

    def complete_task_list(
        self,
        project_id: str,
        task_id_list: List[str],
        target_phase: TaskPhase,
        target_phase_stage: int,
        reply_comment: Optional[str] = None,
        inspection_status: Optional[InspectionStatus] = None,
        parallelism: Optional[int] = None,
    ):
        """
        検査コメントのstatusを変更（対応完了 or 対応不要）にした上で、タスクを受け入れ完了状態にする
        Args:
            project_id: 対象のproject_id
            task_id_list:
            target_phase: 操作対象のタスクフェーズ
            target_phase_stage: 操作対象のタスクのフェーズステージ
            reply_comment: 未回答の検査コメントに対する指摘
            inspection_status: 未処置の検査コメントの状態
        """

        project_title = self.facade.get_project_title(project_id)
        logger.info(f"{project_title} のタスク {len(task_id_list)} 件に対して、今のフェーズを完了状態にします。")

        success_count = 0

        if parallelism is not None:
            partial_func = partial(
                self.complete_task_for_task_wrapper,
                project_id=project_id,
                target_phase=target_phase,
                target_phase_stage=target_phase_stage,
                reply_comment=reply_comment,
                inspection_status=inspection_status,
            )

            with multiprocessing.Pool(parallelism) as pool:
                result_bool_list = pool.map(partial_func, enumerate(task_id_list))
                success_count = len([e for e in result_bool_list if e])

        else:
            # 逐次処理
            for task_index, task_id in enumerate(task_id_list):
                result = self.complete_task(
                    project_id,
                    task_id,
                    task_index=task_index,
                    target_phase=target_phase,
                    target_phase_stage=target_phase_stage,
                    reply_comment=reply_comment,
                    inspection_status=inspection_status,
                )
                if result:
                    success_count += 1

        logger.info(f"{success_count} / {len(task_id_list)} 件のタスクに対して、今のフェーズを完了状態にしました。")


class ComleteTasks(AbstractCommandLineInterface):
    """
    タスクを受け入れ完了にする
    """

    @staticmethod
    def validate(args: argparse.Namespace) -> bool:
        COMMON_MESSAGE = "annofabcli task complete: error:"
        if args.phase == TaskPhase.ANNOTATION.value:
            if args.inspection_status is not None:
                logger.warning(f"'--phase'に'{TaskPhase.ANNOTATION.value}'を指定しているとき、" f"'--inspection_status'の値は無視されます。")
        elif args.phase in [TaskPhase.INSPECTION.value, TaskPhase.ACCEPTANCE.value]:
            if args.reply_comment is not None:
                logger.warning(
                    f"'--phase'に'{TaskPhase.INSPECTION.value}'または'{TaskPhase.ACCEPTANCE.value}'を指定しているとき、"
                    f"'--reply_comment'の値は無視されます。"
                )

        if args.parallelism is not None and not args.yes:
            print(
                f"{COMMON_MESSAGE} argument --parallelism: '--parallelism'を指定するときは、必ず'--yes'を指定してください。",
                file=sys.stderr,
            )
            return False

        return True

    def main(self):
        args = self.args
        if not self.validate(args):
            return

        task_id_list = annofabcli.common.cli.get_list_from_args(args.task_id)
        inspection_status = InspectionStatus(args.inspection_status) if args.inspection_status is not None else None
        project_id = args.project_id
        super().validate_project(project_id, [ProjectMemberRole.OWNER, ProjectMemberRole.ACCEPTER])

        main_obj = CompleteTasksMain(self.service, all_yes=self.all_yes)
        main_obj.complete_task_list(
            project_id,
            task_id_list=task_id_list,
            target_phase=TaskPhase(args.phase),
            target_phase_stage=args.phase_stage,
            inspection_status=inspection_status,
            reply_comment=args.reply_comment,
            parallelism=args.parallelism,
        )


def parse_args(parser: argparse.ArgumentParser):
    argument_parser = ArgumentParser(parser)

    argument_parser.add_project_id()
    argument_parser.add_task_id(required=True)

    parser.add_argument(
        "--phase",
        type=str,
        required=True,
        choices=[TaskPhase.ANNOTATION.value, TaskPhase.INSPECTION.value, TaskPhase.ACCEPTANCE.value],
        help=("操作対象のタスクのフェーズを指定してください。"),
    )

    parser.add_argument(
        "--phase_stage",
        type=int,
        default=1,
        help=("操作対象のタスクのフェーズのステージ番号を指定してください。デフォルトは'1'です。"),
    )

    parser.add_argument(
        "--reply_comment",
        type=str,
        help=(
            f"未回答の検査コメントに対する返信コメントを指定してください。"
            f"'--phase'に'{TaskPhase.ANNOTATION.value}' を指定したときのみ有効なオプションです。"
            f"指定しない場合は、未回答の検査コメントを含むタスクをスキップします。"
        ),
    )

    parser.add_argument(
        "--inspection_status",
        type=str,
        choices=[InspectionStatus.ERROR_CORRECTED.value, InspectionStatus.NO_CORRECTION_REQUIRED.value],
        help=(
            "操作対象のフェーズ未処置の検査コメントをどの状態に変更するかを指定します。"
            f"'--phase'に'{TaskPhase.INSPECTION.value}'または'{TaskPhase.ACCEPTANCE.value}' を指定したときのみ有効なオプションです。"
            "指定しない場合、未処置の検査コメントが含まれるタスクはスキップします。"
            f"{InspectionStatus.ERROR_CORRECTED.value}: 対応完了,"
            f"{InspectionStatus.NO_CORRECTION_REQUIRED.value}: 対応不要"
        ),
    )

    parser.add_argument(
        "--parallelism", type=int, help="使用するプロセス数（並列度）を指定してください。指定する場合は必ず'--yes'を指定してください。指定しない場合は、逐次的に処理します。"
    )

    parser.set_defaults(subcommand_func=main)


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    ComleteTasks(service, facade, args).main()


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "complete"
    subcommand_help = "タスクを完了状態にして次のフェーズに進めます。（教師付の提出、検査/受入の合格）"
    description = (
        "タスクを完了状態にして次のフェーズに進めます。（教師付の提出、検査/受入の合格） "
        "教師付フェーズを完了にする場合は、未回答の検査コメントに対して返信することができます"
        "（未回答の検査コメントに対して返信しないと、タスクを提出できないため）。"
        "検査/受入フェーズを完了する場合は、未処置の検査コメントを対応完了/対応不要状態に変更できます"
        "（未処置の検査コメントが残っている状態では、タスクを合格にできないため）。"
    )
    epilog = "チェッカーまたはオーナロールを持つユーザで実行してください。"

    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description, epilog=epilog)
    parse_args(parser)



import argparse
import logging
import sys
from typing import Any, Dict, List, Optional

from annofabapi.models import JobInfo, JobType, Project

import annofabcli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import AbstractCommandLineInterface, ArgumentParser, build_annofabapi_resource_and_login
from annofabcli.common.enums import FormatArgument

logger = logging.getLogger(__name__)


class ListLastJob(AbstractCommandLineInterface):
    """
    ジョブ一覧を表示する。
    """

    def get_last_job(self, project_id: str, job_type: JobType) -> Optional[JobInfo]:
        """
        最新のジョブを取得する。ジョブが存在しない場合はNoneを返す。

        """
        query_params = {"type": job_type.value}
        content, _ = self.service.api.get_project_job(project_id, query_params)
        job_list = content["list"]
        if len(job_list) == 0:
            logger.debug(f"project_id={project_id}にjob_type={job_type.value}のジョブは存在しませんでした。")
            return None
        else:
            return job_list[-1]

    def get_project_info(self, project_id: str, project: Project, add_details: bool = False) -> Dict[str, Any]:
        """
        出力対象であるプロジェクトに関する情報を取得する。

        Args:
            project_id:
            project: プロジェクト情報
            add_details:

        Returns:
            プロジェクト情報

        """

        project_info = {"project_id": project["project_id"], "project_title": project["title"]}

        if add_details:
            project_info["last_tasks_updated_datetime"] = project["summary"]["last_tasks_updated_datetime"]

            annotation_specs_history = self.service.api.get_annotation_specs_histories(project_id)[0]
            project_info["last_annotation_specs_updated_datetime"] = annotation_specs_history[-1]["updated_datetime"]

        return project_info

    @annofabcli.utils.allow_404_error
    def get_project(self, project_id: str) -> Project:
        project, _ = self.service.api.get_project(project_id)
        return project

    def get_last_job_list(
        self, project_id_list: List[str], job_type: JobType, add_details: bool = False
    ) -> List[JobInfo]:
        job_list = []
        for project_id in project_id_list:
            project = self.get_project(project_id)

            if project is None:
                logger.warning(f"project_id='{project_id}' のプロジェクトは存在しませんでした。")
                continue

            project_info = self.get_project_info(project_id=project_id, add_details=add_details, project=project)
            job = self.get_last_job(project_id, job_type)
            if job is not None:
                job_list.append({**project_info, **job})
            else:
                job_list.append(project_info)

        return job_list

    def print_job_list(self, project_id_list: List[str], job_type: JobType, add_details: bool = False) -> None:
        """
        ジョブ一覧を出力する

        Args:
            project_id: 対象のproject_id
            job_type: ジョブタイプ
        """

        job_list = self.get_last_job_list(project_id_list, job_type=job_type, add_details=add_details)
        logger.info(f"{len(job_list)} 個のプロジェクトの, job_type={job_type.value} の最新ジョブを出力します。")
        self.print_according_to_format(job_list)

    def get_project_id_list(self, organization_name: str) -> List[str]:
        """
        進行中で自分自身が所属する project_id を取得する

        Args:
            organization_name:

        Returns:

        """
        my_account, _ = self.service.api.get_my_account()
        query_params = {"status": "active", "account_id": my_account["account_id"]}
        project_list = self.service.wrapper.get_all_projects_of_organization(
            organization_name, query_params=query_params
        )
        return [e["project_id"] for e in project_list]

    def main(self):
        args = self.args
        job_type = JobType(args.job_type)

        if args.organization is not None:
            project_id_list = self.get_project_id_list(args.organization)

        elif args.project_id is not None:
            project_id_list = annofabcli.common.cli.get_list_from_args(args.project_id)

        else:
            print("引数に`--project_id` または `--organization` を指定してください。", file=sys.stderr)
            return

        self.print_job_list(project_id_list, job_type=job_type, add_details=args.add_details)


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    ListLastJob(service, facade, args).main()


def parse_args(parser: argparse.ArgumentParser):
    argument_parser = ArgumentParser(parser)

    job_choices = [e.value for e in JobType]
    parser.add_argument("--job_type", type=str, choices=job_choices, required=True, help="ジョブタイプを指定します。")

    list_group = parser.add_mutually_exclusive_group(required=True)
    list_group.add_argument(
        "-p",
        "--project_id",
        type=str,
        nargs="+",
        help="対象のプロジェクトのproject_idを指定してください。`file://`を先頭に付けると、一覧が記載されたファイルを指定できます。",
    )

    list_group.add_argument("-org", "--organization", type=str, help="組織配下のすべてのプロジェクトのジョブを出力したい場合は、組織名を指定してください。")

    parser.add_argument(
        "--add_details",
        action="store_true",
        help="プロジェクトに関する詳細情報を表示します" "（`task_last_updated_datetime, annotation_specs_last_updated_datetime`）",
    )

    argument_parser.add_format(
        choices=[FormatArgument.CSV, FormatArgument.JSON, FormatArgument.PRETTY_JSON], default=FormatArgument.CSV
    )
    argument_parser.add_output()
    argument_parser.add_csv_format()

    argument_parser.add_query()
    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "list_last"
    subcommand_help = "複数のプロジェクトに対して、最新のジョブを出力します。"
    description = "複数のプロジェクトに対して、最新のジョブを出力します。"

    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description)
    parse_args(parser)



import csv
from enum import Enum
from typing import List, Tuple

import numpy as np
import pandas as pd


class TimeUnitTarget(Enum):
    H = "h"
    M = "m"
    S = "s"


class FormatTarget(Enum):
    DETAILS = "details"
    TOTAL = "total"
    BY_NAME_TOTAL = "by_name_total"
    COLUMN_LIST = "column_list"
    COLUMN_LIST_PER_PROJECT = "column_list_per_project"


def timeunit_conversion(df: pd.DataFrame, time_unit: TimeUnitTarget = TimeUnitTarget.H) -> pd.DataFrame:
    if time_unit == TimeUnitTarget.H:
        df["worktime_actual"] = df["worktime_actual"] / 60
        df["worktime_monitored"] = df["worktime_monitored"] / 60
        df["worktime_planned"] = df["worktime_planned"] / 60
    elif time_unit == TimeUnitTarget.S:
        df["worktime_actual"] = df["worktime_actual"] * 60
        df["worktime_monitored"] = df["worktime_monitored"] * 60
        df["worktime_planned"] = df["worktime_planned"] * 60

    return df


def calc_df_total(df: pd.DataFrame) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    # 複数のproject_id分を合計
    del df["project_id"]
    del df["project_title"]
    df["user_biography"] = df["user_biography"].fillna("")
    total_df = pd.DataFrame(df.groupby(["user_name", "user_id", "date", "user_biography"], as_index=False).sum())

    # 合計を計算する
    sum_by_date = (
        total_df[["date", "worktime_planned", "worktime_actual", "worktime_monitored"]]
        .groupby(["date"], as_index=False)
        .sum()
    )
    sum_by_name = (
        total_df[["user_name", "worktime_planned", "worktime_actual", "worktime_monitored"]]
        .groupby(["user_name"], as_index=False)
        .sum()
    )
    sum_all = total_df[["worktime_planned", "worktime_actual", "worktime_monitored"]].sum().to_frame().transpose()

    # 比率を追加する
    sum_by_date["activity_rate"] = sum_by_date["worktime_actual"] / sum_by_date["worktime_planned"]
    sum_by_date["monitor_rate"] = sum_by_date["worktime_monitored"] / sum_by_date["worktime_actual"]

    sum_by_name["activity_rate"] = sum_by_name["worktime_actual"] / sum_by_name["worktime_planned"]
    sum_by_name["monitor_rate"] = sum_by_name["worktime_monitored"] / sum_by_name["worktime_actual"]

    sum_all["activity_rate"] = sum_all["worktime_actual"] / sum_all["worktime_planned"]
    sum_all["monitor_rate"] = sum_all["worktime_monitored"] / sum_all["worktime_actual"]

    total_df["activity_rate"] = total_df["worktime_actual"] / total_df["worktime_planned"]
    total_df["monitor_rate"] = total_df["worktime_monitored"] / total_df["worktime_actual"]

    return total_df, sum_by_date, sum_by_name, sum_all


def print_time_list_from_work_time_list(df: pd.DataFrame) -> pd.DataFrame:
    # 複数のproject_id分を合計
    total_df, sum_by_date, sum_by_name, sum_all = calc_df_total(df=df)
    # 不要になったuser_id を削除
    del total_df["user_id"]
    del total_df["user_biography"]
    # 結果を合体する
    result = (
        pd.concat(
            [
                total_df.round(2).replace({np.inf: "--", np.nan: "--"}),
                sum_by_date.round(2).replace({np.inf: "--", np.nan: "--"}),
                sum_by_name.round(2).replace({np.inf: "--", np.nan: "--"}),
                sum_all.round(2).replace({np.inf: "--", np.nan: "--"}),
            ],
            sort=False,
        )
        .sort_values(
            ["date", "user_name"],
        )
        .loc[
            :,
            [
                "date",
                "user_name",
                "worktime_planned",
                "worktime_actual",
                "worktime_monitored",
                "activity_rate",
                "monitor_rate",
            ],
        ]
    )
    # indexをつける
    result = result.set_index(["date", "user_name"])
    # 整形
    result = result.stack(0).unstack(1, fill_value="✕").unstack(1, fill_value="✕")
    result[result.loc[:, pd.IndexSlice[:, ["activity_rate", "monitor_rate"]]] == "✕"] = "--"

    result.index = result.index.date

    # NaT と NaN を 書き換える
    result = result.rename(columns={np.nan: "総合計"}, index={pd.NaT: "合計"})

    return result


def print_byname_total_list(df: pd.DataFrame) -> pd.DataFrame:
    # 複数のproject_id分を合計
    _, _, sum_by_name, _ = calc_df_total(df=df)

    # 結果を合体する
    result = sum_by_name.round(2).replace({np.inf: "--", np.nan: "--"})
    return result


def print_total(df: pd.DataFrame) -> pd.DataFrame:
    # 複数のproject_id分を合計
    _, _, _, sum_all = calc_df_total(df=df)

    # 結果を合体する
    result = sum_all.round(2).replace({np.inf: "--", np.nan: "--"})
    return result


def create_column_list(df: pd.DataFrame) -> pd.DataFrame:
    total_df, _, _, _ = calc_df_total(df=df)

    # 結果を合体する
    result = (
        total_df[
            [
                "date",
                "user_id",
                "user_name",
                "user_biography",
                "worktime_planned",
                "worktime_actual",
                "worktime_monitored",
                "activity_rate",
                "monitor_rate",
            ]
        ]
        .round(2)
        .replace({np.inf: "--", np.nan: "--"})
    )
    return result


def create_column_list_per_project(df: pd.DataFrame) -> pd.DataFrame:
    df["user_biography"] = df["user_biography"].fillna("")
    df["activity_rate"] = df["worktime_actual"] / df["worktime_planned"]
    df["monitor_rate"] = df["worktime_monitored"] / df["worktime_actual"]
    # 出力対象の列を指定する
    result = (
        df[
            [
                "date",
                "project_id",
                "project_title",
                "user_id",
                "user_name",
                "user_biography",
                "worktime_planned",
                "worktime_actual",
                "worktime_monitored",
                "activity_rate",
                "monitor_rate",
                "working_description",
            ]
        ]
        .round(2)
        .replace({np.inf: "--", np.nan: "--"})
    )

    return result


def print_time_list_csv(print_time_list: list) -> None:
    for time_list in print_time_list:
        print(*time_list, sep=",")


def add_id_csv(csv_path: str, id_list: List[str]):
    id_list[0:0] = ["project_id"]

    with open(csv_path, mode="a", encoding="utf_8_sig", newline="\r\n") as f:
        writer = csv.writer(f, delimiter=",")
        writer.writerow([])
        writer.writerow(id_list)



import argparse
import importlib
import logging
from pathlib import Path

import pandas

import annofabcli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import AbstractCommandLineInterface, build_annofabapi_resource_and_login
from annofabcli.common.utils import _catch_exception

logger = logging.getLogger(__name__)


def write_task_histogram(csv: Path, output_dir: Path, minimal_output: bool = False) -> None:
    """
    ヒストグラムを出力する
    """
    task_df = pandas.read_csv(str(csv))
    if len(task_df) == 0:
        logger.warning(f"タスク一覧が0件のため、グラフを出力しません。")
        return

    # holivesのloadに時間がかかって、helpコマンドの出力が遅いため、遅延ロードする
    histogram_module = importlib.import_module("annofabcli.statistics.histogram")
    histogram_obj = histogram_module.Histogram(outdir=str(output_dir))  # type: ignore
    _catch_exception(histogram_obj.write_histogram_for_worktime)(task_df)
    _catch_exception(histogram_obj.write_histogram_for_other)(task_df)

    if not minimal_output:
        _catch_exception(histogram_obj.write_histogram_for_annotation_worktime_by_user)(task_df)
        _catch_exception(histogram_obj.write_histogram_for_inspection_worktime_by_user)(task_df)
        _catch_exception(histogram_obj.write_histogram_for_acceptance_worktime_by_user)(task_df)


class WriteTaskHistogram(AbstractCommandLineInterface):
    def main(self):
        args = self.args
        write_task_histogram(csv=args.csv, output_dir=args.output_dir, minimal_output=args.minimal)


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    WriteTaskHistogram(service, facade, args).main()


def parse_args(parser: argparse.ArgumentParser):
    parser.add_argument(
        "--csv",
        type=Path,
        required=True,
        help=("CSVファイルのパスを指定してください。" "CSVは、'statistics visualize'コマンドの出力結果である'タスクlist.csv'と同じフォーマットです。"),
    )
    parser.add_argument("-o", "--output_dir", type=Path, required=True, help="出力ディレクトリのパス")

    parser.add_argument(
        "--minimal",
        action="store_true",
        help="必要最小限のファイルを出力します。",
    )

    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "write_task_histogram"
    subcommand_help = "CSVからタスク関係のヒストグラムを出力します。"
    description = "CSVからタスク関係のヒストグラムを出力します。"
    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description)
    parse_args(parser)



import argparse
import logging
from typing import Any, Dict, List

import pandas

import annofabcli
import annofabcli.common.cli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import AbstractCommandLineInterface, ArgumentParser, build_annofabapi_resource_and_login
from annofabcli.common.utils import isoduration_to_hour

logger = logging.getLogger(__name__)


class LaborTimePerUser(AbstractCommandLineInterface):
    """
    メンバ別の作業時間を出力する。
    """

    def get_labor_time_per_user(self, project_id: str) -> List[Dict[str, Any]]:
        """
        メンバ別の作業時間をCSVに出力するための dict 配列を作成する。

        Args:
            project_id:

        Returns:
            メンバ別の作業時間をCSVに出力するための dict 配列

        """
        account_statistics = self.service.wrapper.get_account_statistics(project_id)
        row_list: List[Dict[str, Any]] = []
        for stat_by_user in account_statistics:
            account_id = stat_by_user["account_id"]
            member = self.facade.get_project_member_from_account_id(project_id, account_id)

            histories = stat_by_user["histories"]
            for stat in histories:
                stat["account_id"] = account_id
                stat["user_id"] = member["user_id"] if member is not None else None
                stat["username"] = member["username"] if member is not None else None
                stat["biography"] = member["biography"] if member is not None else None
                stat["worktime_hour"] = isoduration_to_hour(stat["worktime"])

            row_list.extend(histories)
        return row_list

    def list_cumulative_labor_time(self, project_id: str) -> None:
        super().validate_project(project_id, project_member_roles=None)

        account_stat_list = self.get_labor_time_per_user(project_id)
        df = pandas.DataFrame(account_stat_list)
        if len(df) > 0:
            # 出力対象の列を指定する
            target_df = df[
                [
                    "date",
                    "account_id",
                    "user_id",
                    "username",
                    "biography",
                    "worktime_hour",
                    "tasks_completed",
                    "tasks_rejected",
                ]
            ]
            annofabcli.utils.print_csv(target_df, output=self.output, to_csv_kwargs=self.csv_format)
        else:
            logger.error(f"出力対象データが0件のため、出力しません。")

    def main(self):
        args = self.args

        project_id = args.project_id
        self.list_cumulative_labor_time(project_id)


def parse_args(parser: argparse.ArgumentParser):
    argument_parser = ArgumentParser(parser)

    argument_parser.add_project_id()
    argument_parser.add_csv_format()
    argument_parser.add_output()

    parser.set_defaults(subcommand_func=main)


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    LaborTimePerUser(service, facade, args).main()


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "list_labor_time_per_user"
    subcommand_help = "メンバ別の作業時間、完了数、差し戻し回数を出力する。"
    description = "メンバ別の作業時間、完了数、差し戻し回数をCSV形式で出力する。"
    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description=description)
    parse_args(parser)



import argparse
import json
import logging
import sys
from pathlib import Path
from typing import Dict, List, Optional

from annofabapi.models import InputData, TaskStatus
from annofabapi.parser import SimpleAnnotationParser

import annofabcli
import annofabcli.common.cli
from annofabcli.common.cli import ArgumentParser
from annofabcli.common.image import IsParserFunc, write_annotation_images_from_path
from annofabcli.common.typing import InputDataSize

logger = logging.getLogger(__name__)


class WriteAnnotationImage:
    @staticmethod
    def create_is_target_parser_func(
        task_status_complete: bool = False, task_id_list: Optional[List[str]] = None
    ) -> IsParserFunc:
        def is_target_parser(parser: SimpleAnnotationParser) -> bool:
            simple_annotation = parser.parse()
            if task_status_complete:
                if simple_annotation.task_status != TaskStatus.COMPLETE:
                    logger.debug(
                        f"task_statusがcompleteでない( {simple_annotation.task_status.value})ため、"
                        f"{simple_annotation.task_id}, {simple_annotation.input_data_name} はスキップします。"
                    )
                    return False

            if task_id_list is not None and len(task_id_list) > 0:
                if simple_annotation.task_id not in task_id_list:
                    logger.debug(
                        f"画像化対象外のタスク {simple_annotation.task_id} であるため、 {simple_annotation.input_data_name} はスキップします。"
                    )
                    return False

            return True

        return is_target_parser

    COMMON_MESSAGE = "annofabcli filesystem write_annotation_image: error:"

    @staticmethod
    def get_input_data_dict(input_data_json: Path) -> Dict[str, InputData]:
        with input_data_json.open() as f:
            logger.debug(f"{input_data_json} を読み込み中")
            input_data_list = json.load(f)
            logger.debug(f"{len(input_data_list)} 件をDictに変換中")
            return {e["input_data_id"]: e for e in input_data_list}

    def main(self, args):
        logger.info(f"args: {args}")

        image_size: Optional[InputDataSize] = None
        if args.image_size is not None:
            image_size = annofabcli.common.cli.get_input_data_size(args.image_size)
            if image_size is None:
                logger.error("--image_size のフォーマットが不正です")
                print(
                    f"{self.COMMON_MESSAGE} argument --image_size: フォーマットが不正です。",
                    file=sys.stderr,
                )
                return

        if args.input_data_json is not None:
            if args.metadata_key_of_image_size is None:
                print(
                    f"{self.COMMON_MESSAGE} argument --metadata_key_of_image_size: "
                    f"`--input_data_json`を指定した場合、`--metadata_key_of_image_size`は必須です。",
                    file=sys.stderr,
                )
                return

        # label_color_dict を取得する
        label_color_dict = annofabcli.common.cli.get_json_from_args(args.label_color)
        label_color_dict = {k: tuple(v) for k, v in label_color_dict.items()}
        label_name_list = (
            annofabcli.common.cli.get_list_from_args(args.label_name) if args.label_name is not None else None
        )
        annotation_path: Path = args.annotation
        task_id_list = annofabcli.common.cli.get_list_from_args(args.task_id)
        is_target_parser_func = self.create_is_target_parser_func(args.task_status_complete, task_id_list)

        input_data_dict = self.get_input_data_dict(args.input_data_json) if args.input_data_json is not None else None

        # 画像生成
        result = write_annotation_images_from_path(
            annotation_path,
            label_color_dict=label_color_dict,
            output_dir_path=Path(args.output_dir),
            image_size=image_size,
            input_data_dict=input_data_dict,
            metadata_key_of_image_width=args.metadata_key_of_image_size[0]
            if args.metadata_key_of_image_size is not None
            else None,
            metadata_key_of_image_height=args.metadata_key_of_image_size[1]
            if args.metadata_key_of_image_size is not None
            else None,
            output_image_extension=args.image_extension,
            background_color=args.background_color,
            label_name_list=label_name_list,
            is_target_parser_func=is_target_parser_func,
        )
        if not result:
            logger.error(f"'{annotation_path}' のアノテーション情報の画像化に失敗しました。")


def main(args):
    WriteAnnotationImage().main(args)


def parse_args(parser: argparse.ArgumentParser):
    argument_parser = ArgumentParser(parser)

    parser.add_argument("--annotation", type=Path, required=True, help="アノテーションzip、またはzipを展開したディレクトリ")

    image_size_group = parser.add_mutually_exclusive_group(required=True)

    image_size_group.add_argument("--image_size", type=str, help="画像サイズ。{width}x{height}。ex) 1280x720")

    image_size_group.add_argument(
        "--metadata_key_of_image_size",
        type=str,
        nargs=2,
        help="画像サイズが設定された、入力データのメタデータのキー。" "`--image_size_key_of_metadata {width_key} {height_key}`",
    )

    parser.add_argument(
        "--input_data_json",
        type=Path,
        help="入力データ情報が記載されたJSONファイルのパスを指定してください。引数`--metadata_key_of_image_size`を指定したときに必須です。"
        "JSONファイルは`$ annofabcli project download input_data`コマンドで取得できます。",
    )

    parser.add_argument(
        "--label_name",
        type=str,
        nargs="+",
        required=False,
        help="画像化対象のlabel_nameを指定します。指定しない場合は、すべてのlabel_nameが画像化対象になります。"
        "`file://`を先頭に付けると、label_name の一覧が記載されたファイルを指定できます。",
    )

    parser.add_argument(
        "--label_color",
        type=str,
        required=True,
        help='label_nameとRGBの関係をJSON形式で指定します。ex) `{"dog":[255,128,64], "cat":[0,0,255]}`'
        "`file://`を先頭に付けると、JSON形式のファイルを指定できます。",
    )

    parser.add_argument("-o", "--output_dir", type=str, required=True, help="出力ディレクトリのパス")

    parser.add_argument("--image_extension", type=str, default="png", help="出力画像の拡張子")

    parser.add_argument(
        "--background_color",
        type=str,
        help=(
            "アノテーションの画像の背景色を指定します。"
            'ex):  "rgb(173, 216, 230)", "lightgrey",  "#add8e6" '
            "[ImageColor Module](https://hhsprings.bitbucket.io/docs/programming/examples/python/PIL/ImageColor.html) がサポートする文字列を利用できます。"  # noqa: E501
            "指定しない場合は、黒（rgb(0,0,0)）になります。"
        ),
    )

    parser.add_argument("--task_status_complete", action="store_true", help="taskのstatusがcompleteの場合のみ画像を生成します。")

    argument_parser.add_task_id(
        required=False,
        help_message=(
            "画像化するタスクのtask_idを指定します。" "指定しない場合、すべてのタスクが画像化されます。" "`file://`を先頭に付けると、task_idの一覧が記載されたファイルを指定できます。"
        ),
    )

    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "write_annotation_image"

    subcommand_help = "アノテーションzipを展開したディレクトリから、アノテーションの画像（Semantic Segmentation用）を生成する。"

    description = "アノテーションzipを展開したディレクトリから、アノテーションの画像（Semantic Segmentation用）を生成する。" + "矩形、ポリゴン、塗りつぶし、塗りつぶしv2が対象。"

    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description)
    parse_args(parser)
